/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package campus;


import serp.util.Options;
import java.io.BufferedReader;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.InputStreamReader;


/**
 *  A tracer that outputs raw log messages from STDIN or a log file,
 *  roughly respecting the timestamp.  It can be a standalone application
 *  or an embedded engine.  The first field of input is expected to be
 *  timestamp in seconds and the field separator is tab.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: LogTracer.java,v 1.2 2003/09/16 03:41:39 glchen Exp $
 */
public class LogTracer 
{
    /** An observer who takes tracer's output. */
    public interface TracerObserver
    {
        /** feed the output to the observer. */
        public void handleMessage(String msg)
            throws Exception;
    }
    
    private BufferedReader _in;
    private TracerObserver _observer;
    private long _clock = 0;        // corresponding to clock in input
    private long _tstamp = 0;       // actual execution in milliseconds


    //=======================================================================
    // Constructors.
    //=======================================================================

    public LogTracer(Options opts)
        throws Exception
    {
        String log = opts.getProperty("log");
        InputStream in = (log==null) ? System.in : new FileInputStream(log);
        _in = new BufferedReader(new InputStreamReader(in), 1);
    }

    //=======================================================================
    // Set tracer observer.
    //=======================================================================

    public void setObserver(TracerObserver ob)
    {
        _observer = ob;
    }

    public void run()
        throws Exception
    {
        String line = _in.readLine();
        while (line != null)
        {
            long ts = Long.parseLong(line.substring(0, 10));
            if (_clock == 0)
            {
                // first line ever
                _clock = ts;
                _tstamp = System.currentTimeMillis();
                processOneLine(line);
            }
            else
            {
                if (_clock == ts)
                {
                    // consecutive lines with same clock (in seconds)
                    processOneLine(line);
                }
                else
                {
                    long sleep = (ts - _clock) * 1000 - 
                        (System.currentTimeMillis() - _tstamp); // working time
                    if (sleep > 10) Thread.sleep(sleep);
                    _clock = ts;
                    _tstamp = System.currentTimeMillis();
                    processOneLine(line);
                }
            }
            line = _in.readLine();
        }
    }

    private void processOneLine(String line)
        throws Exception
    {
        if (_observer != null)
        {
            _observer.handleMessage(line);
        }
        else
        {
            System.out.println(line);
        }
    }


    //=======================================================================
    // Main entrance.
    //=======================================================================

    public static void main(String[] args)
        throws Exception
    {
        Options opts = new Options();
        opts.setFromCmdLine(args);
        new LogTracer(opts).run();
    }
}
