/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package campus.syslog;


import serp.util.Options; 
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.List;
import java.util.Vector;

import solar.Client;
import solar.api.Event;
import solar.api.Attribute;
import solar.api.Request;
import solar.impl.EventFactory;
import solar.service.dan.Source;
import campus.syslog.SyslogConstants;


/**
 *  A syslog source that produces events from syslog messages.
 *  It parses the syslog msg stream, and produces syslog events. 
 *  It generates and publishes an event every time it receives
 *  a new event from the server.
 *  
 *  <p>The published events contain following attributes:
 *  <ul>
 *  <li>1. timestamp    (when the event is received)
 *  <li>2. year         (year of the timestamp)
 *  <li>3. month        (year of the timestamp)
 *  <li>4. day          (year of the timestamp)
 *  <li>5. time         (year of the timestamp)
 *  <li>6. msgtype      (type of the message get from syslog)
 *  <li>7. apname       (name of Access point)
 *  <li>4. clienthost   (client hostname or IP address)
 *  <li>5. clientmac    (client's MAC address)
 *  <li>6. action       (such as authorized, roamed, etc...)
 *  <li>7. note         (explanation for the action)
 *  </ul>
 *
 *  @author     Guanling Chen
 *  @author     Yurong Xu
 *  @version    $Id: SyslogSource.java,v 1.5 2003/09/19 00:38:24 glchen Exp $
 */

public class SyslogSource
    extends Source 
    implements SyslogConstants
{
    private boolean _standalone;

    //=======================================================================
    // Constructors.
    //=======================================================================

    public SyslogSource(boolean standalone)
    {
        _standalone = standalone;
    }


    //=======================================================================
    // Implementation of solar.service.dan.Source interface.
    //=======================================================================
    
    public Object handleQuery(Request query)
        throws Exception
    {
        return null;
    }

    public void run()
        throws Exception
    {
        BufferedReader in = 
            new BufferedReader(new InputStreamReader(System.in), 1);
        // parse the input and publishes event
        String line = in.readLine();
        while(line != null)
        {
            String[] fields = line.split("\t");
            try
            {
                List attr = new Vector();
                int idx = 0;
                attr.add(new Attribute(EVT_TIMESTAMP, fields[idx++]));
                attr.add(new Attribute(EVT_YEAR, fields[idx++]));
                attr.add(new Attribute(EVT_MONTH, fields[idx++]));
                attr.add(new Attribute(EVT_DAY, fields[idx++]));
                attr.add(new Attribute(EVT_TIME, fields[idx++]));
                attr.add(new Attribute(EVT_MSGTYPE, fields[idx++]));
                attr.add(new Attribute(EVT_APNAME, fields[idx++]));
                attr.add(new Attribute(EVT_CLIENTHOST, fields[idx++]));
                attr.add(new Attribute(EVT_CLIENTMAC, fields[idx++]));
                attr.add(new Attribute(EVT_ACTION, fields[idx++]));
                attr.add(new Attribute(EVT_NOTE, fields[idx++]));
                Event evt = EventFactory.getPlainEvent(attr);
                // Publish the event
                if (_standalone)
                {
                    System.out.println(evt);
                    System.out.flush();
                }
                else
                {
                    publish(evt);
                }
            }
            catch (Exception ex)
            {
                System.err.println(line);
                ex.printStackTrace();
            }
            line = in.readLine();
        }
    }


    /**
     *  Read the configuration from command line and start the source
     *  as Syslog client or a standalone program.  The recognized options
     *  are:
     *  
     *  <ul>
     *  <ui>-standalone: whether to run the source without Solar.<BR>
     *  <ui>-host: the host address of the connecting Planet.<BR>
     *  <ui>-port: the port number of the connecting Planet.<BR>
     *  </ul>
     *  
     *  The Solar name assigned for this source is:
     *  [source=syslog, wlan=campus]
     */
    public static void main(String args[])
        throws Exception
    {
        Options opts = new Options();
        opts.setFromCmdLine(args);

        boolean standalone = opts.getBooleanProperty("standalone", false);
        SyslogSource src = new SyslogSource(standalone);
        
        // if standalone, run directly
        if (standalone)
        {
            src.run();
            return;
        }

        // otherwise, connect to specified planet and make advertisement
        String host = opts.getProperty("host","tahoe");
        int port = opts.getIntProperty("port", 5470);

        String name = "[source=syslog, wlan=campus]";        
        Client client = new Client(host, port);
        client.advertise(name, src);
    }   
}
