/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package campus.sysnort.trigger;


import serp.util.Options;
import java.io.PrintWriter;
import java.io.FileOutputStream;
import java.util.Date;
import java.util.Iterator;
import java.util.Hashtable;

import campus.syslog.SyslogConstants;
import campus.sysnort.Trigger;
import solar.api.Event;
import solar.util.SystemUtil;


/**
 *  Monitors the number of clients on the APs and update the hotspot
 *  plots periodically.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: HotspotTrigger.java,v 1.4 2003/09/08 23:50:35 glchen Exp $
 */
public class HotspotTrigger 
    extends Trigger
    implements SyslogConstants
{
    protected Hashtable _aps;         //(apname, counter)
    protected Hashtable _clients;     //(clientmac, apname)

    protected int _plotRate;        // plotting interval in seconds
    protected int _hotThreshold;    // alert threshold on number of clients
    protected String _datafile;     // the name of hotspot data file
    protected String _plotpath;     // the path to place plots and datafile
    protected String _makefile;     // the path to sysnort makefile
    protected boolean _blocking;    // blocking call to command of plotting
    
    protected long _clock;
    protected Process _process;

    
    //=======================================================================
    // Implementation of campus.sysnort.Trigger interface.
    //=======================================================================

    public void init(Options opts)
        throws Exception
    {
        _aps = new Hashtable();
        _clients = new Hashtable();
        _hotThreshold = opts.getIntProperty("hot_threshold", 10);
        _plotRate = opts.getIntProperty("plot_rate", 30);  // default 30 sec

        _datafile = opts.getProperty("data_file", "hotspot.txt");
        _plotpath = opts.getProperty("plot_path", 
            "/u/glchen/public_html/sysnort/");
        _makefile = opts.getProperty("makefile", 
            "/u/glchen/work/solar/script/Makefile.sysnort");
        _blocking = opts.getBooleanProperty("blocking", false);
    }
    
    public boolean checkCondition(Event evt)
        throws Exception
    {
        boolean ret = false;
        long timestamp = Long.parseLong((String)evt.getValue(EVT_TIMESTAMP));
        String clientmac = (String) evt.getValue(EVT_CLIENTMAC);
        String apname = (String) evt.getValue(EVT_APNAME);
        String action = (String) evt.getValue(EVT_ACTION);
        if (clientmac == null || apname == null)
        {
            return false;
        }
        if (action.equals(TAG_ASSOCIATED) 
            || action.equals(TAG_REASSOCIATED))
        {
            String oldap = (String) _clients.get(clientmac);
            if (oldap != null)
            {
                updateCounter(oldap, -1);
            }
            int cnt = updateCounter(apname, +1);
            _clients.put(clientmac, apname);
            // is this a really hot AP?
            if (cnt > _hotThreshold)
            {
                String msg = new Integer(cnt) + " " +
                    apname + " " + new Date(timestamp*1000);
                _actionObject = msg;
                ret = true;
            }
        }
        else if (action.equals(TAG_DEAUTHENTICATING)
            || action.equals(TAG_DEAUTHENTICATION)
            || action.equals(TAG_DEASSOCIATING)
            || action.equals(TAG_DEASSOCIATION))
        {
            String oldap = (String) _clients.get(clientmac);
            if (oldap != null)
            {
                updateCounter(oldap, -1);
                _clients.remove(clientmac);
            }
        }
        // determine whether need to plot the hotspots (periodically)
        if (_clock == 0)
        {
            _clock = timestamp;
        }
        else
        {
            if (timestamp - _clock > _plotRate)
            {
                _clock = timestamp;
                ret = true;
            }
        }
        return ret;
    }

    private int updateCounter(String apname, int delta)
    {
        int ret;
        Integer c = (Integer) _aps.get(apname);
        if (c == null)
        {
            ret = delta;
        }
        else
        {
            ret = c.intValue() + delta;
        }
        _aps.put(apname, new Integer(ret));
        return ret;
    }

    public void takeAction(Object actionObject)
        throws Exception
    {
        if (actionObject != null)
        {
            // info about the hotspot
            System.out.println(getName() + ": " + actionObject);
        }
        boolean plot = false;
        // prepare the datafile for plots
        PrintWriter pw = new PrintWriter(
            new FileOutputStream(_plotpath + _datafile));
        Iterator itr = _aps.keySet().iterator();
        while (itr.hasNext())
        {
            String apname = (String) itr.next();
            Integer cnt = (Integer) _aps.get(apname);
            if (cnt.intValue() > 0)
            {
                pw.println(apname + " " + _aps.get(apname));
            }
            if (!plot && cnt.intValue() > 1)
            {
                plot = true;
            }
        }
        pw.flush();
        pw.close();

        if (_process != null) _process.destroy();
        if (plot)
        {
            _process = SystemUtil.command(
                "make -f " + _makefile + " hotspot",
                _plotpath);
            if (_blocking) _process.waitFor();
        }
    }
}
