/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */
 
package rice.pastry.socket;


import java.io.*;
import java.nio.*;
import java.nio.channels.*;
import java.nio.charset.*;
import java.util.*;
import java.net.*;

import rice.pastry.*;
import rice.pastry.dist.*;
import rice.pastry.routing.*;
import rice.pastry.messaging.*;
import rice.pastry.wire.exception.*;
import rice.pastry.wire.messaging.datagram.*;
import rice.pastry.wire.messaging.socket.*;

import solar.api.*;
import solar.impl.*;
import solar.service.transport.*;


/**
 *  A direct mimic of WireNodeHandle class.
 *  Class which represents a node handle in the socket-based pastry protocol.
 *  Initially, all of the messages are sent over UDP.  If a message is too large
 *  to be sent over the UDP protocol (as determined by the MAX_UDP_MESSAGE_SIZE),
 *  then a socket connection is opened to the remote node.
 *
 *  @author  Guanling Chen
 *  @author  Alan Mislove
 *  @version $Id: SocketNodeHandle.java,v 1.4 2003/10/05 18:50:53 glchen Exp $
 */
public class SocketNodeHandle 
    extends DistNodeHandle
    implements TransportListener
{
    // the throttle (pings only sent this many seconds)
    public static int PING_THROTTLE = 5;
    // the time the last ping was performed
    protected transient long lastpingtime;
    
    protected transient int _protocol;
    protected transient EventFactory _factory;
    protected transient boolean _greeted;
    protected transient Object _lock = new Object();
    protected transient List _attr = new Vector();

    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public SocketNodeHandle(int proto, InetSocketAddress address, NodeId nid) 
    {
        super(nid, address);
        debug("creating Socket handle for node: " + nid + " address: " + address);
        lastpingtime = 0;
        _protocol = proto;
        _factory = new EventFactory();
    }

    public SocketNodeHandle(int proto, InetSocketAddress address, NodeId nid, PastryNode pn) 
    {
        this(proto, address, nid);
        setLocalNode(pn);
    }


    //=======================================================================
    // Implementation of Transport.TransportListener interface.
    //=======================================================================
    
//    public void dataReceived(byte[] data)
    public void dataReceived(Adu adu, Header header)
    {
        try
        {
//            Event evt = (Event) _factory.parse(data);
            Event evt = (Event) adu;
            Transport transport = ((SocketPastryNode)getLocalNode()).getTransport();
            SocketPastryNode localNode = (SocketPastryNode)getLocalNode();
            Object obj = evt.getValue("message");
            debug("SocketNodeHandle: " + obj);
            if (obj instanceof NodeIdRequestMessage)
            {
                InetSocketAddress dest = (InetSocketAddress) evt.getValue("address");
                _attr.clear();
                _attr.add(new Attribute("message", new NodeIdResponseMessage(nodeId)));
                transport.send(
                    EventFactory.getSerializedEvent(_attr).toBytes(),
                    new solar.service.transport.SocketAddress(dest));
            }
            else if (obj instanceof rice.pastry.socket.HelloMessage)
            {
                debug("********** got hello *********");
                rice.pastry.socket.HelloMessage hm = 
                    (rice.pastry.socket.HelloMessage) obj;
                SocketNodeHandle handle = 
                    ((SocketNodeHandlePool) localNode.getNodeHandlePool()).get(hm.getNodeId());
                if (handle == null)
                {
                    handle = new SocketNodeHandle(_protocol,
                        hm.getAddress(), hm.getNodeId(), localNode);
                    handle = (SocketNodeHandle) localNode.getNodeHandlePool().coalesce(handle);
                }
                _attr.clear();
                _attr.add(new Attribute("message", hm.getResponse(localNode)));
                transport.send(
                    EventFactory.getSerializedEvent(_attr).toBytes(),
                    new solar.service.transport.SocketAddress(hm.getAddress()));
                debug("********* sent hello response *********");
            }
            else if (obj instanceof rice.pastry.socket.HelloResponseMessage)
            {
                /*
                synchronized (_lock)
                {
                    _lock.notifyAll();
                }*/
                debug("********* got hello response *********");
            }
            else if (obj instanceof SocketTransportMessage)
            {
                SocketTransportMessage stm = (SocketTransportMessage) obj;
                receiveMessageImpl((Message)stm.getObject());
            }
            else
            {
                System.err.println(
                    "Unknown message for SocketNodeHandle: " + obj);
            }
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
            debug(ex.toString());
        }
    }

    public void connectionOpened(solar.api.Address peer)
    {
        
    }

    public void connectionClosed(solar.api.Address peer)
    {
        
    }
    

    //=======================================================================
    // Implementation of DistNodeHandle interface.
    //=======================================================================

    /**
     * Method which is called when a SocketCommandMessage comes across an open
     * socket for this node handle.
     *
     * @param message The message coming across the wire.
     */
    public void receiveSocketMessage(SocketCommandMessage message) 
    {
        System.err.println("receiveSocketMessage: " + message);
    }

    /**
     * Called to send a message to the node corresponding to this handle.
     *
     * @param msg Message to be delivered, may or may not be routeMessage.
     */
    public void receiveMessageImpl(Message msg) 
    {
        assertLocalNode();
        SocketPastryNode spn = (SocketPastryNode) getLocalNode();

        if (isLocal) 
        {
            debug("Sending message " + msg + " locally");
            spn.receiveMessage(msg);
            return;
        }
        
        debug("Passing message " + msg + " for socket writing");

        try
        {
            Transport transport = ((SocketPastryNode)getLocalNode()).getTransport();
            List attr = new Vector();
//            synchronized (_lock)
//            {
                if (!_greeted)
                {
                    attr.add(new Attribute("message", 
                        new HelloMessage((SocketPastryNode) getLocalNode(), nodeId)));
                    transport.send(
                        EventFactory.getSerializedEvent(attr).toBytes(),
                        new solar.service.transport.SocketAddress(address));
//                    _lock.wait();
                    _greeted = true;
//                    debug("********* greeting received *********");
                }
//            }
            attr.clear();
            attr.add(new Attribute("message", new SocketTransportMessage(msg, nodeId)));
            transport.send(
                EventFactory.getSerializedEvent(attr).toBytes(),
                new solar.service.transport.SocketAddress(address));
//            Thread.sleep(100);
        }
        catch (Exception ex)
        {
//            ex.printStackTrace();
            debug(ex.toString());
        }
    }

    public boolean pingImpl() 
    {
        if (isLocal) 
        {
            setProximity(0);
            return alive;
        }

        if (getLocalNode() != null) 
        {
            if (((SocketNodeHandle) getLocalNode().getLocalHandle()).getAddress().getAddress().equals(address.getAddress())) 
            {
                setProximity(1);
                return alive;
            }
        }
      
        long now = System.currentTimeMillis();
        if (now - lastpingtime < PING_THROTTLE*1000)
        return alive;

        lastpingtime = now;

        if (getLocalNode() != null) 
        {
            // always send ping over UDP
            //((WirePastryNode) getLocalNode()).getDatagramManager().write(nodeId, address, new PingMessage(getLocalNode().getNodeId(), nodeId, 0, this));
        }

        return alive;
//        return true;
    }


    public String toStringImpl() 
    {
        return nodeId + " (" + address.getAddress().getHostAddress() + ":" + address.getPort() + ")";
    }
}
