/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar;


import serp.util.Options;
import org.apache.log4j.Logger;
import org.apache.log4j.Level;
import org.apache.log4j.BasicConfigurator;
import java.io.FileInputStream;
import java.util.List;

import solar.api.Guid;
import solar.service.SolarServices;
import solar.service.ServiceManager;
import solar.api.Event;
import solar.service.fap.FapClientService;
import solar.service.directory.Name;
import solar.service.dan.Source;
import solar.service.dan.Application;
import solar.service.dan.DanClientService;
import solar.service.dispatch.DispatchService;
import solar.util.FileUtil;


/**
 *  The facade class to interact with Solar system.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: Client.java,v 1.30 2003/09/26 19:47:14 glchen Exp $
 */
public class Client 
{
    private static Logger _log = Logger.getLogger("solar");

    private DispatchService _dispatch;
    private FapClientService _fap;
    private DanClientService _dan;

    
    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public Client(String host, int port)
        throws Exception
    {
        Options opts = getDefaultOptions();
        opts.setProperty("service.dan.planet", 
            host + ":" + Integer.toString(port));
        construct(opts);
    }

    public Client(String host, int port, Options opts)
        throws Exception
    {
        String serialize = opts.getProperty("serialize");
        if (serialize == null)
        {
            serialize = opts.getProperty("service.edge_transport.serialize");
        }
        if (opts.getProperty("cfg") == null)
        {
            opts = getDefaultOptions();
        }
        opts.setProperty("service.edge_transport.serialize", serialize);
        opts.setProperty("service.dan.planet", 
            host + ":" + Integer.toString(port));
        construct(opts);
    }

    public Client(String cfg)
        throws Exception
    {
        Options opts = new Options();
        opts.load(new FileInputStream(cfg));
        construct(opts);
    }

    public Client(Options opts)
        throws Exception
    {
        construct(opts);
    }

    private Options getDefaultOptions()
    {
        Options opts = new Options();
        opts.setProperty("service.dispatch.classname", 
            "solar.service.dispatch.SolarDispatchService");
        opts.setProperty("service.dispatch.transport", "edge_transport");

        opts.setProperty("service.dan.classname", 
            "solar.service.dan.DanClientService");
        opts.setProperty("service.dan.rpc", "edge_rpc");
        opts.setProperty("service.dan.transport", "edge_transport");

        opts.setProperty("service.fap.classname",
            "solar.service.fap.FapClientService");
            
        opts.setProperty("service.edge_rpc.classname",
            "solar.service.rpc.SolarRpcService");
        opts.setProperty("service.edge_rpc.transport", "edge_transport");
            
        opts.setProperty("service.edge_transport.classname", 
            "solar.service.transport.UdpTransportService");
        opts.setProperty("service.edge_transport.port", 0);
        return opts;
    }

    /** Common operations needed by constructors. */
    private void construct(Options opts)
        throws Exception
    {
        // set the logging verbose level
        boolean debug = opts.getBooleanProperty("debug", false);
        if (debug)
        {
            _log.setLevel( (Level) Level.DEBUG );
        }
        else
        {
            _log.setLevel( (Level) Level.INFO );
        }
        BasicConfigurator.configure();

        // init service manager and relavant services
        ServiceManager sm = SolarServices.getInstance();
        sm.init(opts);
        _dispatch = (DispatchService) sm.getService("dispatch");
        _fap = (FapClientService) sm.getService("fap");
        _dan = (DanClientService) sm.getService("dan");
    }


    //=======================================================================
    // Public interface methods.
    //=======================================================================

    /** Advertise a static name for a Solar source. */
    public Guid advertise(String name, Source src)
        throws Exception
    {
        return _dan.advertise(name, "nullobj", src);
    }

    /** Advertise a static name for a Solar source. */
    public Guid advertise(String name, Object obj, Source src)
        throws Exception
    {
        return _dan.advertise(name, obj, src);
    }
    
    /** Advertise a dynamic name for a Solar source. */
    public Guid advertise(String graph, String name, Object obj, Source src)
        throws Exception
    {
        return _dan.advertise(graph, name, obj, src);
    }

    /** Return a list of names and name records matching the name query. */
    public List lookup(String query)
        throws Exception
    {
        return _dan.lookup(query);
    }

    /** Subscribe to any source whose name matches the query. */
    public Guid subscribe(String query, Application app)
        throws Exception
    {
        return _dan.subscribe(query, app);
    }

    /** Compose an operator graph and subscribe to one of the operators. */
    public Guid subscribe(String graph, String root, Application app)
        throws Exception
    {
        return _dan.subscribe(graph, root, app);
    }

    
    //=======================================================================
    // Main entrance.
    //=======================================================================

    /**
     *  Start the client and make name lookups. Command line
     *  usage:
     *  <p>
     *  -host tahoe -port 5470 -lookup "[source=clock]"
     */
    public static void main(String[] args)
        throws Exception
    {
        Options opts = new Options();
        opts.setFromCmdLine(args);

        String host = opts.getProperty("host", "hogback");
        int port = opts.getIntProperty("port", 5470);
        String query = opts.getProperty("lookup");
        
        Client cl = new Client(host, port);
        System.out.println(cl.lookup(query));
    }

}
