/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.impl;


import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.util.List;
import java.util.Vector;
import java.util.Iterator;

import solar.api.Event;
import solar.api.Attribute;
import solar.api.ParsingException;


/**
 *  SerializedEvent containts attributes whose value are arbituary objects
 *  and will be serialized into byte array before transmission.
 *
 *  @author     Guanling Chen
 *  @version    $Id: SerializedEvent.java,v 1.6 2003/09/23 14:03:57 glchen Exp $
 */
public class SerializedEvent
    extends AbstractAttributeRecord
    implements Event
{
    //=======================================================================
    // Constructors.
    //=======================================================================

    /** Constructs an event with a list of attributes. */
    public SerializedEvent(List aList)
    {
        _attributes = new Vector(aList);
    }


    /** Constructs an event by copying from another one. */
    public SerializedEvent(Event aEvent)
    {
        _attributes = new Vector(aEvent.getAttributes());
    }

    
    /** Constructs an event by parsing an encoded byte array. */
    public SerializedEvent(byte[] cnt)
        throws ParsingException
    {
        this(cnt, 0, cnt.length);
    }


    /** Constructs an event by parsing an encoded byte array. */
    public SerializedEvent(byte[] cnt, int offset, int len)
        throws ParsingException
    {
        _attributes = parseSerializedEvent(cnt, offset, len);
    }


    //=======================================================================
    // Private helper method.
    //=======================================================================

    private List parseSerializedEvent(byte[] cnt, int offset, int len)
        throws ParsingException
    {
        List ret = new Vector();
        try
        {
            // construct the object input stream from the byte array
            ByteArrayInputStream bytes = 
                new ByteArrayInputStream(cnt, offset, len);
            ObjectInputStream in = new ObjectInputStream(bytes);

            // read the number of attributes and the sequence of 
            // key-value pairs
            int size = in.readInt();
            for (int i=0; i<size; i++)
            {
                String key = (String) in.readObject();
                Object val = in.readObject();
                ret.add(new Attribute(key, val));
            }
        }
        catch (Exception ioe)
        {
            throw new ParsingException("Fail to parse the byte array.", ioe);
        }
        return ret;
    }

    
    //=======================================================================
    // Implementation of solar.api.Event interface.
    //=======================================================================
    
    public int getEncoding()
    {
        return EVENT_SERIALIZED;
    }


    //=======================================================================
    // Implementation of solar.api.AttributeRecord interface.
    //=======================================================================
    
    public boolean equals(Object obj)
    {
        if (obj == null || !(obj instanceof SerializedEvent))
            return false;

        SerializedEvent other = (SerializedEvent) obj;
        return _attributes.equals(other._attributes);
    }
    

    //=======================================================================
    // Implementation of solar.api.Adu interface.
    //=======================================================================
    
    public String toString()
    {
        return _attributes.toString();
    }
    
    public byte[] toBytes()
    {
        try
        {
            // construct object output stream from a new byte array
            // output stream
            ByteArrayOutputStream bytes = new ByteArrayOutputStream();
            ObjectOutputStream out = new ObjectOutputStream(bytes);

            // write the number of attributes and the sequence of attributes
            // in key-value pairs
            out.writeInt(_attributes.size());
            Iterator itr = _attributes.iterator();
            while (itr.hasNext())
            {
                Attribute attr = (Attribute) itr.next();
                out.writeObject(attr.getKey());
                out.writeObject(attr.getValue());
            }

            // get the final byte array and add prefix of one byte as
            // the event encoding type, so the receiver of this byte
            // array knows it is a serialized event
            byte[] se = bytes.toByteArray();
            byte[] ret = new byte[se.length+1];
            ret[0] = EVENT_SERIALIZED;
            System.arraycopy(se, 0, ret, 1, se.length);
            return ret;
        }
        catch (IOException ioe)
        {
            ioe.printStackTrace();
            return null;
        }
    }
}
