/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.dan;


import serp.util.Options;
import java.util.List;

import solar.api.Guid;
import solar.api.Request;
import solar.api.Address;
import solar.api.ParsingException;
import solar.service.BaseService;
import solar.service.ServiceException;
import solar.service.dan.comm.*;
import solar.service.dan.lang.*;
import solar.service.fap.DataSource;
import solar.service.fap.DataSink;
import solar.service.fap.InputPort;
import solar.service.fap.OutputPort;
import solar.service.fap.FapService;
import solar.service.fap.FapConstants;
import solar.service.directory.Name;
import solar.service.directory.PlainName;
import solar.service.directory.NameRecord;
import solar.service.directory.DirectoryService;
import solar.service.directory.DirectoryException;
import solar.service.rpc.RpcService;
import solar.service.rpc.RpcException;
import solar.service.transport.SocketAddress;
import solar.service.transport.TransportService;


/**
 *  The abstract implementation (mostly dealing with clients) of DAN (Data
 *  Aggregation Network) service.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: AbstractDanService.java,v 1.4 2003/10/05 18:50:54 glchen Exp $
 */
public abstract class AbstractDanService 
    extends BaseService
    implements DanService, DanConstants, FapConstants
{
    protected FapService _fap;
    protected DirectoryService _directory;
    
    protected RpcService _coreRpc;
    protected RpcService _edgeRpc;

    protected TransportService _coreTransport;
    protected TransportService _edgeTransport;

    protected Address _laddr;
    protected PipeConnector _pc;
    protected NodeManager _nodeman = new NodeManager();
    protected SpecParser _parser; // graph parser
    
    
    //=======================================================================
    // Implementation of solar.service.Service interface.
    //=======================================================================

    public void init()
        throws ServiceException
    {
    }
    
    public void initDependents()
        throws ServiceException
    {
        super.initDependents();     // must call at the top

        _fap = (FapService) 
            getDependentService("fap", "fap");

        _directory = (DirectoryService) 
            getDependentService("directory", "directory");

        _edgeRpc = (RpcService) 
            getDependentService("edgeRpc", "edge_rpc");
        _edgeTransport = (TransportService) 
            getDependentService("edgeTransport", "edge_transport");
        
        _coreRpc = (RpcService) 
            getDependentService("coreRpc", "core_rpc");
        _coreTransport = (TransportService) 
            getDependentService("coreTransport", "core_transport");
                
        _laddr = _coreTransport.getLocalAddress();

        _nodeman = new NodeManager();
        _parser = new XmlSpecParser();
        _pc = new PipeConnector(this);
    }

    public Object handleRequest(Request request)
      throws Exception
    {
        Object ret = null;
        if (request instanceof AdvertiseRequest)
        {
            ret = advertise((AdvertiseRequest)request);
        }
        else if (request instanceof LookupRequest)
        {
            ret = lookup((LookupRequest)request);
        }
        else if (request instanceof SubscribeRequest)
        {
            ret = subscribe((SubscribeRequest)request);
        }
        else if (request instanceof PipeConnectRequest)
        {
            ret = _pc.pconnect((PipeConnectRequest)request);
        }
        else
        {
            throw new RpcException(
                "Failed to handle unknown DAN request " + request + ".");
        }
        return ret;
    }


    //=======================================================================
    // Implementation of solar.service.dan.DanService interface.
    //=======================================================================

    public Address getLocalAddress()
    {
        return _laddr;
    }

    public FapService getFapService()
    {
        return _fap;
    }
    
    public RpcService getRpcService()
    {
        return _coreRpc;
    }

    public DirectoryService getDirectoryService()
    {
        return _directory;
    }

    public NodeManager getNodeManager()
    {
        return _nodeman;
    }
    
    public Guid makeGuid()
    {
        return Guid.makeGuid(_laddr);
    }

    protected GraphManager makeGraphManager(GraphSpec gs)
    {
        return new LocalGraphManager(gs);
    }

    
    //=======================================================================
    // Private helper methods.
    //=======================================================================

    private List lookup(LookupRequest request)
        throws Exception
    {
        Name query = new PlainName((String) request.getQuery());
        return _directory.lookup(query);
    }
    
    private Guid advertise(AdvertiseRequest request)
        throws Exception
    {
        GraphSpec gs = null;
        String root = null;
        // get request arguments
        String graph = request.getGraphSpec();
        Name name = new PlainName((String) request.getName());
        if (graph != null)
        {
            gs = _parser.parse(graph);
            root = GraphSpec.getOpVariable(name);
            if (gs.getFilter(root) == null)
            {
                throw new ParsingException(
                    "Operator in CS name not defined in the graph.");
            }
        }

        Object extra = request.getNameExtra();
        Address client = request.getSourceAddress();
        Guid nid = request.getGuid();

        //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        // TODO: fix the advertisement interface for DanClientService
//        String outport = (String) req.getArgument(DAN_OUT_PORT);
//        outport = (outport==null) ? DAN_OUTPORT : outport;
        String outport = DAN_OUTPORT;
        
        // create a relay proxy for client, and register it
        NodeHandler proxy = new NodeHandler(new RelayOperator());
        Guid guid = makeGuid();
        _nodeman.putLocal(guid, proxy);

        // create a monitor that makes advertisement, and register it
        Monitor mon = new Monitor(this);
        Guid monid = makeGuid();
        _nodeman.putLocal(monid, new NodeHandler(new ProxyOperator(mon)));

        // load the graph if there is any, and subscribe the monitor
        if (gs != null)
        {
            connect(FAP_PUSH, 
                loadGraph(gs, root), outport, 
                monid, DAN_INPORT);
        }

        // make the static or dynamic advertisement
        mon.advertise(guid, name, extra);
        
        return guid;
    }
    
    private Guid subscribe(SubscribeRequest request)
        throws Exception
    {
        // get request arguments
        GraphSpec gs = _parser.parse(request.getGraphSpec());
        String root = request.getGraphRoot();
        Guid nid = request.getGuid();
        Address client = request.getSourceAddress();

        //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        // TODO: fix the subscribe interface for DanClientService
//        String outport = (String) req.getArgument(DAN_OUT_PORT);
//        outport = (outport==null) ? DAN_OUTPORT : outport;
        String outport = DAN_OUTPORT;
        
        // create a relay proxy for client, and register it
        NodeHandler proxy = new NodeHandler(new RelayOperator());
        Guid guid = makeGuid();
        _nodeman.putLocal(guid, proxy);
        
        // load the graph and setup the pipe from root to proxy
        connect(FAP_PUSH,
            loadGraph(gs, root), outport, 
            guid, DAN_INPORT);
        return guid;
    }

    private Guid loadGraph(GraphSpec gs, String op)
        throws Exception
    {
        GraphManager gm = makeGraphManager(gs);
        gm.init(this);
        return gm.loadOperator(op);
    }
}
