/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.dan;


import serp.util.Options;
import java.util.List;
import java.util.Iterator;

import solar.api.Guid;
import solar.api.Request;
import solar.api.Address;
import solar.service.BaseService;
import solar.service.ServiceException;
import solar.service.dan.comm.LookupRequest;
import solar.service.dan.comm.SubscribeRequest;
import solar.service.dan.comm.AdvertiseRequest;
import solar.service.directory.Name;
import solar.service.directory.DirectoryService;
import solar.service.fap.DataSink;
import solar.service.fap.DataSource;
import solar.service.fap.InputPort;
import solar.service.fap.OutputPort;
import solar.service.fap.FapService;
import solar.service.fap.FapConstants;
import solar.service.rpc.RpcService;
import solar.service.rpc.RpcConstants;
import solar.service.rpc.RpcException;
import solar.service.transport.SocketAddress;
import solar.service.transport.TransportService;


/**
 *  The client interface for DAN (Data Aggregation Network) service.
 *  
 *  We enforce that all the pipe setup requests involving a Solar client are 
 *  always issued from that client. The Planet never issues request to setup 
 *  a pipe to a Solar client. So we can be sure that the remote end for a 
 *  requesting pipe is always on the Planet.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: DanClientService.java,v 1.3 2003/10/05 18:50:54 glchen Exp $
 */
public class DanClientService 
    extends BaseService
    implements DanService, RpcConstants, DanConstants, FapConstants
{
    private Address _planet;
    private Address _laddr;
    
    protected FapService _fap;
    protected RpcService _rpc;
    protected TransportService _transport;

    protected PipeConnector _pc;
    protected NodeManager _nodeman;
    

    //=======================================================================
    // Implementation of solar.service.Service interface.
    //=======================================================================
    
    public void init()
        throws ServiceException
    {
        try
        {
            Options config = getServiceManager().getServiceConfig(getName());
            _planet = new SocketAddress(config.getProperty("planet"));
        }
        catch (ServiceException se)
        {
            throw se;
        }
        catch (Exception ex)
        {
            throw new ServiceException(ex);
        }
    }

    public void initDependents()
        throws ServiceException
    {
        super.initDependents();     // must call at the top

        _fap = (FapService) 
            getDependentService("fap", "fap");
        _rpc = (RpcService) 
            getDependentService("edge_rpc", "edge_rpc");
        _transport = (TransportService) 
            getDependentService("edge_transport", "edge_transport");
        _laddr = _transport.getLocalAddress();

        _nodeman = new NodeManager();
        _pc = new PipeConnector(this);
    }


    //=======================================================================
    // Implementation of solar.service.dan.DanService interface.
    //=======================================================================

    public void connect(int type, 
        Guid sguid, String outport, Guid dguid, String inport)
        throws DanException
    {
        // do not need in dan client service.
    }

    public Address getLocalAddress()
    {
        return _laddr;
    }

    public FapService getFapService()
    {
        return _fap;
    }
    
    public RpcService getRpcService()
    {
        return _rpc;
    }

    public DirectoryService getDirectoryService()
    {
        return null;    // do not need in dan client service.
    }

    public NodeManager getNodeManager()
    {
        return _nodeman;
    }
    
    public Guid makeGuid()
    {
        return Guid.makeGuid(_laddr);
    }


    //=======================================================================
    // Public lookup-related methods.
    //=======================================================================

    /** Resolve a name query in the name space. */
    public List lookup(String query)
        throws DanException
    {
        Request request = new LookupRequest(query);
        request.setSourceAddress(_laddr);
        try
        {
            return (List) callRpc(request);
        }
        catch (Exception ex)
        {
            throw new DanException("Failed to make name lookup.", ex);
        }
    }

    //=======================================================================
    // Public advertise-related methods.
    //=======================================================================

    /** Advertise a static name with a name object and return an ID. */
    public Guid advertise(String name, Object extra, Source src)
        throws DanException
    {
        return advertise(null, name, extra, src);
    }

    /** Advertise a dynamic name with a name object and return an ID. */
    public Guid advertise(String graph, String name, Object extra, Source src)
        throws DanException
    {   
        // prepare RPC request
        Guid guid = makeGuid();
        Request request = new AdvertiseRequest(graph, name, extra, guid);
        request.setSourceAddress(_laddr);
        try
        {
            // get the remote proxy ID from RPC call
            Guid rpid = (Guid) callRpc(request);
            // create a handler to run the source
            NodeHandler handler = new NodeHandler(src);
            // register the source
            _nodeman.putLocal(guid, handler);
            // set up the mixed pipe from local source to remote sink
            _pc.pconnect(FAP_MIXED, 
                guid, DAN_OUTPORT,
                _planet, rpid, DAN_INPORT,
                true);
            // start the source
            handler.start();
            // return the source ID
            return guid;
        }
        catch (Exception ex)
        {
            throw new DanException("Failed to make advertisement.", ex);
        }
    }

    //=======================================================================
    // Public subscribe-related methods.
    //=======================================================================

    /** Subscribe to any source whose name matches specified query. */
    public Guid subscribe(String query, Application app)
        throws DanException
    {
        String graph = "<graph>\n";
        graph += "\t<filter name=\"$root\"\n";
        graph += "\t\tclassname=\"solar.service.dan.RelayOperator\"\n";
        graph += "\t\tcodebase=\"http://localhost/\">\n";
        graph += "\t</filter>\n";
        graph += "\t<pipe type=\"push\" name=\"p\" source=\"" 
            + query + "\"\n";
        graph += "\t\tsink=\"$root\"/>\n";
        graph += "</graph>\n";
        return subscribe(graph, "$root", app);
    }

    /** Subscribe to a specified operator defined in a graph. */
    public Guid subscribe(String graph, String root, Application app)
        throws DanException
    {
        // prepare RPC request
        Guid guid = makeGuid();
        Request request = new SubscribeRequest(graph, root, guid);
        request.setSourceAddress(_laddr);
        try
        {
            // get the remote proxy ID as the return of RPC
            Guid rpid = (Guid) callRpc(request);
            // create a node handler as the data sink
            _nodeman.putLocal(guid, new NodeHandler(app));
            // set up the pipe from remote source to local sink/app
            _pc.pconnect(FAP_PUSH, 
                _planet, rpid, DAN_OUTPORT,
                guid, DAN_INPORT, 
                true);
            // return the app ID
            return guid;
        }
        catch (Exception ex)
        {
            throw new DanException("Failed to make subscription.", ex);
        }
    }

    //=======================================================================
    // Private helper methods.
    //=======================================================================

    /** Invoke a RPC call with specified request. */
    private Object callRpc(Request req)
        throws RpcException
    {
        return _rpc.invoke(_planet, getName(), req, RPC_TIMEOUT);
    }
}
