/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.dan;


import serp.util.Options;
import java.util.List;
import java.util.Vector;

import solar.api.Id;
import solar.api.Event;
import solar.api.Request;
import solar.api.Identifiable;


/**
 *  The interface for all nodes in the graph. There are three types of
 *  nodes: sources, operators, and applications.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: GraphNode.java,v 1.2 2003/09/25 05:29:56 glchen Exp $
 */
public abstract class GraphNode
    implements Identifiable, DanConstants
{
    protected Id _gid;
    private NodeHandler _handler;
    
    
    //=======================================================================
    // Implementation of solar.service.api.Identifiable.
    //=======================================================================
    
    public void setId(Id gid)
    {
        if (_gid != null)
        {
            throw new IllegalStateException(
                "Cannot set the ID more than once.");
        }
        _gid = gid;
    }

    public Id getId()
    {
        return _gid;
    }


    //=======================================================================
    // Public interface methods.
    //=======================================================================

    /** Customized initialization derived from graph spec. */
    public void init(Options opts)
        throws Exception
    {
        // default as empty, overload if needed.
    }

    
    /** Initialize the source with a handler. */
    public void init(NodeHandler handler)
    {
        if (_handler != null)
        {
            throw new IllegalStateException(
                "Graph node has already been initialized.");
        }
        _handler = handler;
    }


    /** Return a list of input ports of this node. */
    public List getInputPorts()
    {
        List ret = new Vector();
        ret.add(DAN_INPORT); // default to have only one input port
        return ret;
    }


    /** Return a list of output ports of this node. */
    public List getOutputPorts()
    {
        List ret = new Vector();
        ret.add(DAN_OUTPORT); // default to have only one output port
        return ret;
    }

    
    /** Handle the event being actively pushed/published to this operator. */
    public abstract void handleEvent(Event evt)
        throws Exception;


    /** Handle the event coming from the specified input port. */
    public void handleEvent(Event evt, String inport)
        throws Exception
    {
        handleEvent(evt);   // default to ignore input port
    }


    /** Handle the query being actively pulled by downstream operators. */
    public abstract Object handleQuery(Request query)
        throws Exception;


    /** Handle the query coming from the specified output port. */
    public Object handleQuery(Request query, String outport)
        throws Exception
    {
        return handleQuery(query);  // default to ignore output port
    }


    //=======================================================================
    // Private helper methods.
    //=======================================================================
    
    protected final void publish(Event evt)
        throws Exception
    {
        publish(evt, DAN_OUTPORT);
    }

    protected final void publish(Event evt, String outport)
        throws Exception
    {
        if (_handler == null)
        {
            throw new IllegalStateException(
                "Graph node has not been initialized.");
        }
        _handler.publish(evt, outport);
    }
}
