/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.dan;


import java.util.List;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Enumeration;

import solar.api.Guid;
import solar.api.Event;
import solar.api.EventHandler;
import solar.api.Attribute;
import solar.api.AttributeRecord;
import solar.impl.PlainAttributeRecord;
import solar.service.fap.FapConstants;
import solar.service.directory.Name;
import solar.service.directory.NameRecord;
import solar.service.directory.Advertisement;
import solar.service.directory.DirectoryListener;
import solar.service.directory.PlainName;
import solar.service.directory.DirectoryService;


/**
 *  A monitor is attached to some operator to manage its subscriptions or
 *  advertisement.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: Monitor.java,v 1.3 2003/09/25 21:37:37 glchen Exp $
 */
class Monitor 
    implements EventHandler, DirectoryListener, DanConstants, FapConstants
{
    private Guid _guid;   // guid of the adertiser or subscriber
    
    private Name _name;     // name to be advertised
    private Name _pname;    // previously advertised name
    private Object _object;    // the object go with name advertisement

    private Name _query;
    private Name _pquery;

    private DanService _dan;
    private NodeManager _nodeman;
    private DirectoryService _directory;

    
    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public Monitor(DanService dan)
    {
        _dan = dan;
        _nodeman = dan.getNodeManager();
        _directory = dan.getDirectoryService();
    }


    //=======================================================================
    // Public interface methods.
    //=======================================================================

    // TODO: static name can be done more efficiently
    public void advertise(Guid guid, Name name, Object obj)
        throws Exception
    {
        _guid = guid;
        _name = name;
        _object = obj;
        _pname = composeName(name, null);
        _directory.advertise(_pname, guid, _object);
    }

    public void subscribe(Guid guid, Name query)
        throws Exception
    {
        _guid = guid;
        _query = query;
        _directory.addListener((Name)query, this);
        select();
    }


    //=======================================================================
    // Implementation of solar.service.directory.DirectoryListener interface.
    //=======================================================================
    
    public void nameAdded(Name name, NameRecord rec)
    {
        select();
    }

    public void nameRemoved(Name name, NameRecord rec)
    {
        select();        
    }

    public void nameUpdated(Name oldName, 
        NameRecord oldrec, Name newName, NameRecord newrec)
    {
        select();
    }

    public void nameMigrated(Name name, NameRecord oldrec, NameRecord newrec)
    {
        select();
    }


    //=======================================================================
    // Private helper methods.
    //=======================================================================

    private void select()
    {
        try
        {
            Iterator itr = _directory.lookup(_query).iterator();
            if (itr.hasNext())
            {
                Advertisement adv = (Advertisement) itr.next();
                NameRecord nr = adv.getNameRecord();
                Guid guid = (Guid) nr.getId();
                _nodeman.putRemote(guid, nr.getAddress());
                //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                // TODO: use default now, to fix the subscribe interface
                _dan.connect(FAP_PUSH, 
                    guid, DAN_OUTPORT, 
                    _guid, DAN_INPORT);
                _pname = adv.getName();
            }
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }
    
    private Name composeName(Name name, Event evt)
    {
        if (evt == null)
            return name;

        return new PlainName(replaceValue((AttributeRecord)name, evt));
    }

    // replace the values in the CS name with values in the event
    // notice we assume there is only one operator used in CS name
    // so all the events from handleEvent are from this operator
    // and we ignore the operator variable in the CS name and directly
    // use whatever follows the ":". for example, $locator:building_name
    private AttributeRecord replaceValue(AttributeRecord ar, Event evt)
    {
        List ret = new Vector();
        Iterator itr = ar.getAttributes().iterator();
        while (itr.hasNext())
        {
            Attribute attr = (Attribute) itr.next();
            String key = attr.getKey();
            Object val = attr.getValue();
            if (val instanceof AttributeRecord)
            {
                // recursively replacement
                ret.add(new Attribute(key, 
                    replaceValue((AttributeRecord)val, evt)));
            }
            else
            {
                // get the desired value in the event
                String vs = (String) val;
                String[] ks = (vs.substring(vs.indexOf(":") + 1)).split(".");
                try
                {
                    for (int i=0; i<ks.length-1; i++)
                    {
                        // recursively go to last attribute
                        attr = (Attribute) evt.getValue(ks[i]);
                    }
                    if (ks.length > 0)  // indeed a CS name
                    {
                        // value of last attribute has to be string
                        vs = (String) attr.getValue();
                    }
                }
                catch (Exception ex)
                {
                    // if desired attribute not in event
                    vs = "NAME_NOT_MATCH_ERROR";
                }
                ret.add(new Attribute(key, vs));
            }
        }
        return new PlainAttributeRecord(ret);
    }
    

    //=======================================================================
    // Implementation of solar.api.EventHandler interface.
    //=======================================================================

    public void handleEvent(Event evt)
        throws Exception
    {
        if (_name != null)
        {
            Name name = composeName(_name, evt);
            if (!_pname.equals(name))
            {
                _directory.update(_pname, _guid, name, _object);
                _pname = name;
            }
        }
        else
        {
            Name query = new PlainName(
                replaceValue( (AttributeRecord) _query, evt) );
            if (!_pquery.equals(query))
            {
                _directory.addListener((Name)query, this);
                select();
            }
        }
    }
}
