/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.directory;


import serp.util.Options;
import org.apache.log4j.Logger;
import java.io.Serializable;
import java.util.List;
import java.util.Iterator;
import java.util.Hashtable;

import solar.api.Id;
import solar.api.Event;
import solar.api.Request;
import solar.api.Address;
import solar.service.BaseService;
import solar.service.ServiceException;
import solar.service.dispatch.DispatchService;
import solar.service.directory.comm.*;
import solar.service.rpc.RpcService;
import solar.service.rpc.RpcConstants;
import solar.service.rpc.RpcException;
import solar.service.transport.SocketAddress;
import solar.service.transport.TransportService;


/**
 *  A centralized directory service, which requires a rpc and a transport 
 *  services.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: CentralDirectoryService.java,v 1.9 2003/10/05 18:50:56 glchen Exp $
 */
public class CentralDirectoryService 
    extends BaseService
    implements DirectoryService, RpcConstants
{
    /** The logging facility. */
    protected Logger _log = Logger.getLogger(CentralDirectoryService.class);

    protected boolean _local;   // whether this node is the directory server
    protected Address _server;  // the address of the central directory server
    
    protected DispatchService _dispatch;
    protected RpcService _rpc;
    protected String _transport;
    protected Address _laddr;   // the local address of this node

    // instance of name space if this node is the server
    protected Directory _directory;

    // list of listeners on this node while the server is a remote node
    protected Hashtable _listeners;
    
    
    //=======================================================================
    //  Implementation of solar.service.Service interface.
    //=======================================================================
    
    public void init()
        throws ServiceException
    {
        String addr = "";
        try
        {
            // decides whether this node acts as server or client from config
            Options config = getServiceManager().getServiceConfig(getName());
            _transport = config.getProperty("transport", "transport");
            addr = config.getProperty("server");
            if (addr == null || addr.length() == 0)
            {
                _server = _laddr;
                _local = true;
            }
            else
            {
                _server = new SocketAddress(addr);
                _local = (_server.equals(_laddr)) ? true : false;
            }
        }
        catch (ServiceException se)
        {
            throw se;
        }
        catch (Exception ex)
        {
            throw new ServiceException(ex);
        }

        // create a local directory if this node is the central server
        if (_local)
        {
            _directory = new FlatDirectory();
        }
        else
        {
            _listeners = new Hashtable();
        }
    }

    public void initDependents()
        throws ServiceException
    {
        super.initDependents();     // must call at the top

        // get disptach service and local transport address
        _dispatch = (DispatchService) 
            getDependentService("dispatch", "dispatch");
        _laddr = _dispatch.getTransportAddress(_transport);
        // get RPC service
        _rpc = (RpcService) getDependentService("rpc", "rpc");
    }

    public Object handleRequest(Request request)
      throws Exception
    {
        Object ret = null;
        if (request instanceof AddListenerRequest)
        {
            AddListenerRequest alr = (AddListenerRequest) request;
            doAddListener(alr.getPattern(), alr.getListener());
        }
        else if (request instanceof RemoveListenerRequest)
        {
            RemoveListenerRequest rlr = (RemoveListenerRequest) request;
            doRemoveListener(rlr.getPattern(), rlr.getListener());
        }
        else if (request instanceof LookupRequest)
        {
            LookupRequest lr = (LookupRequest) request;
            ret = doLookup(lr.getPattern());
        }
        else if (request instanceof AdvertiseRequest)
        {
            AdvertiseRequest ar = (AdvertiseRequest) request;
            ret = doAdvertise(ar.getName(), ar.getNameRecord());
        }
        else if (request instanceof UnadvertiseRequest)
        {
            UnadvertiseRequest ur = (UnadvertiseRequest) request;
            ret = doUnadvertise(ur.getName(), ur.getNameRecord());
        }
        else if (request instanceof UpdateRequest)
        {
            UpdateRequest ur = (UpdateRequest) request;
            ret = doUpdate(
                ur.getOldName(), ur.getOldNameRecord(), 
                ur.getNewName(), ur.getNewNameRecord());
        }
        else if (request instanceof NameAddedRequest)
        {
            NameAddedRequest nar = (NameAddedRequest) request;
            Id id = nar.getId();
            Name name = nar.getName();
            NameRecord rec = nar.getNameRecord();
            RemoteDirectoryListener rno = 
                (RemoteDirectoryListener) _listeners.get(id);
            rno.setLogger(_log);
            rno.setRpcService(getName(), _rpc);
            if (rno != null) rno.nameAdded(name, rec);
            else throw new DirectoryException("Name listener not found.");
        }
        else if (request instanceof NameRemovedRequest)
        {
            NameRemovedRequest nrr = (NameRemovedRequest) request;
            Id id = nrr.getId();
            Name name = nrr.getName();
            NameRecord rec = nrr.getNameRecord();
            RemoteDirectoryListener rno = 
                (RemoteDirectoryListener) _listeners.get(id);
            rno.setLogger(_log);
            rno.setRpcService(getName(), _rpc);
            if (rno != null) rno.nameRemoved(name, rec);
            else throw new DirectoryException("Name listener not found.");
        }
        else if (request instanceof NameUpdatedRequest)
        {
            NameUpdatedRequest nur = (NameUpdatedRequest) request;
            Id id = nur.getId();
            Name oldname = nur.getOldName();
            Name newname = nur.getNewName();
            NameRecord oldrec = nur.getOldNameRecord();
            NameRecord newrec = nur.getNewNameRecord();
            RemoteDirectoryListener rno = 
                (RemoteDirectoryListener) _listeners.get(id);
            rno.setLogger(_log);
            rno.setRpcService(getName(), _rpc);
            if (rno != null) rno.nameUpdated(oldname, oldrec, newname, newrec);
            else throw new DirectoryException("Name listener not found.");
        }
        else if (request instanceof NameMigratedRequest)
        {
            NameMigratedRequest nmr = (NameMigratedRequest) request;
            Id id = nmr.getId();
            Name name = nmr.getName();
            NameRecord oldrec = nmr.getOldNameRecord();
            NameRecord newrec = nmr.getNewNameRecord();
            RemoteDirectoryListener rno = 
                (RemoteDirectoryListener) _listeners.get(id);
            rno.setLogger(_log);
            rno.setRpcService(getName(), _rpc);
            if (rno != null) rno.nameMigrated(name, oldrec, newrec);
            else throw new DirectoryException("Name listener not found.");
        }
        else
        {
            throw new RpcException(
                "Failed to handle directory request " + request + ".");
        }
        return ret;
    }


    //=======================================================================
    //  Implementation of solar.service.directory.DirectoryService interface.
    //=======================================================================

    public void addListener(Name pattern, DirectoryListener listener)
        throws DirectoryException
    {
        doAddListener(pattern, listener);
    }

    public void removeListener(Name pattern, DirectoryListener listener)
        throws DirectoryException
    {
        doRemoveListener(pattern, listener);        
    }

    public List lookup(Name pattern)
        throws DirectoryException
    {
        return doLookup(pattern);
    }

    public Object advertise(Name name, Id guid, Object obj)
        throws DirectoryException
    {
        return doAdvertise(name, new NameRecord(guid, _laddr, obj));
    }

    public Object unadvertise(Name name, Id guid)
        throws DirectoryException
    {
        return doUnadvertise(name, new NameRecord(guid, _laddr));
    }

    public Object update(Name oldname, Id guid, Name newname, Object obj)
        throws DirectoryException
    {
        return doUpdate(oldname, new NameRecord(guid, _laddr), 
            newname, new NameRecord(guid, _laddr, obj));
    }


    //=======================================================================
    //  Private helper methods.
    //=======================================================================

    private Object callRpc(Request req)
        throws RpcException
    {
        return _rpc.invoke(_server, getName(), req, RPC_TIMEOUT);
    }
    
    private synchronized void doAddListener(Name pattern, 
        DirectoryListener listener)
        throws DirectoryException
    {
        if (_local)
        {
            _directory.addListener(pattern, listener);
        }
        else
        {
            // add listener to locally maintained table
            Id id = Id.makeId("directory-listener-");
            _listeners.put(id, listener);

            // create a remote proxy and tell server to add it
            RemoteDirectoryListener rno = 
                new RemoteDirectoryListener(id, _laddr);
            
            Request req = new AddListenerRequest(pattern, rno);
            try
            {
                callRpc(req);
            }
            catch (RpcException re)
            {
                throw new DirectoryException("Failed to add an listener.", re);
            }
        }
    }

    private synchronized void doRemoveListener(Name pattern, 
        DirectoryListener listener)
        throws DirectoryException
    {
        if (_local)
        {
            _directory.removeListener(pattern, listener);
        }
        else
        {
            // remove listener from locally maintained table
            Id oid = null;
            Iterator itr = _listeners.keySet().iterator();
            while (itr.hasNext())
            {
                Id id = (Id) itr.next();
                if (_listeners.get(id) == listener) // reference comparison
                {
                    oid = id;
                    break;
                }
            }
            if (oid != null)
            {
                _listeners.remove(oid);
            }
            else
            {
                throw new DirectoryException(
                    "The listener to remove does not exist.");
            }
            
            // create a remote proxy and tell server to remove it
            RemoteDirectoryListener rno = 
                new RemoteDirectoryListener(oid, _laddr);
            
            Request req = new RemoveListenerRequest(pattern, rno);
            try
            {
                callRpc(req);
            }
            catch (RpcException re)
            {
                throw new DirectoryException(
                    "Failed to remove an listener.", re);
            }
        }
    }

    private synchronized List doLookup(Name pattern)
        throws DirectoryException
    {
        if (_local)
        {
            return _directory.lookup(pattern);
        }
        else
        {
            Request req = new LookupRequest(pattern);
            try
            {
                return (List) callRpc(req);
            }
            catch (RpcException re)
            {
                throw new DirectoryException("Failed to lookup names.", re);
            }
        }
    }
    
    private synchronized Object doAdvertise(Name name, NameRecord rec)
        throws DirectoryException
    {
        Object ret = null;
        if (_local)
        {
            NameRecord or = _directory.advertise(name, rec);
            ret = (or == null) ? null : or.getObject();
        }
        else
        {
            Request req = new AdvertiseRequest(name, rec);
            try
            {
                ret = callRpc(req);
            }
            catch (RpcException re)
            {
                throw new DirectoryException(
                    "Failed to advertise a name.", re);
            }
        }
        return ret;
    }

    private synchronized Object doUnadvertise(Name name, NameRecord rec)
        throws DirectoryException
    {
        Object ret = null;
        if (_local)
        {
            NameRecord or = _directory.unadvertise(name, rec);
            ret = (or == null) ? null : or.getObject();
        }
        else
        {
            Request req = new UnadvertiseRequest(name, rec);
            try
            {
                ret = callRpc(req);
            }
            catch (RpcException re)
            {
                throw new DirectoryException(
                    "Failed to unadvertise a name.", re);
            }
        }
        return ret;
    }

    private synchronized Object doUpdate(Name oldname, 
        NameRecord oldrec, Name newname, NameRecord newrec)
        throws DirectoryException
    {
        Object ret = null;
        if (_local)
        {
            NameRecord or = _directory.update(oldname, oldrec, newname, newrec);
            ret = (or == null) ? null : or.getObject();
        }
        else
        {
            Request req = new UpdateRequest(oldname, oldrec, newname, newrec);
            try
            {
                ret = callRpc(req);
            }
            catch (RpcException re)
            {
                throw new DirectoryException("Failed to update a name.", re);
            }
        }
        return ret;
    }
}
