/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */


package solar.service.directory;


import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;


/**
 *  A node in the directory tree, either an internal node or a leaf node.
 *  A name node may have a set of children name nodes and a set of name
 *  records (such as network addresses).
 *  
 *  @author     Guanling Chen
 *  @version    $Id: NameNode.java,v 1.1 2003/08/01 20:06:23 glchen Exp $
 */
class NameNode 
{
    protected String _label;
    protected Set _children;
    protected Set _records;
    protected NameNode _parent;

    
    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public NameNode(String label)
    {
        _label = label;
        _children = new HashSet();
        _records = new HashSet();
    }


    //=======================================================================
    // Public interface methods.
    //=======================================================================

    /** Return whether this name node is a leaf of name tree. */
    public boolean isLeaf()
    {
        return (_children.size() == 0);
    }

    /** Return whether this name node is the root of name tree. */
    public boolean isRoot()
    {
        return (_parent == null);
    }

    /** Return the label of this name node. */
    public String getLabel()
    {
        return _label;
    }
    
    /** Set the parent node of this name node. */
    public void setParent(NameNode parent)
    {
        _parent = parent;
    }

    /** Return the parent node of this name node. */
    public NameNode getParent()
    {
        return _parent;
    }

    /** Add a child, either another name node or a name record. */
    public void addChild(NameNode child)
    {
        child.setParent(this);
        _children.add(child);
    }

    /** Create a new child with specified label. */
    public NameNode createChild(String label)
    {
        NameNode ret = new NameNode(label);
        ret.setParent(this);
        this.addChild(ret);
        return ret;
    }

    /** Return a child node with specified label. */
    public NameNode getChild(String label)
    {
        Iterator itr = _children.iterator();
        while (itr.hasNext())
        {
            NameNode c = (NameNode) itr.next();
            if (c.getLabel().equals(label))
                return c;
        }
        return null;
    }

    /** Return the set of children of this name node. */
    public Set getChildren()
    {
        return _children;
    }

    /** Add a name record to this name node. */
    public void addRecord(NameRecord rec)
    {
        _records.add(rec);
    }

    /** Remove a name record from this name node. */
    public void removeRecord(NameRecord rec)
    {
        _records.remove(rec);
    }

    /** Return the set of name records for this name node. */
    public Set getRecords()
    {
        return _records;
    }
}
