/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.fork;


import serp.util.Options;
import org.apache.log4j.Logger;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Hashtable;

import solar.api.Id;
import solar.api.Event;
import solar.api.Request;
import solar.api.Address;
import solar.service.BaseService;
import solar.service.ServiceException;
import solar.service.dispatch.DispatchService;
import solar.service.fork.comm.StartTaskRequest;
import solar.service.fork.comm.TaskFinishEvent;
import solar.service.rpc.RpcService;
import solar.service.rpc.RpcConstants;
import solar.util.SystemUtil;


/**
 *  An implementation of forking service. It reads a configuration file
 *  and sends requests to fork services on other hosts.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: SolarForkService.java,v 1.3 2003/10/08 22:53:22 glchen Exp $
 */
public class SolarForkService 
    extends BaseService
    implements ForkService, TaskBatch.BatchListener, RpcConstants
{
    protected static Logger _log = Logger.getLogger(SolarForkService.class);
    
    protected String _forkName;
    protected String _transportName;
    protected Address _transportAddress;
    protected RpcService _rpc;
    protected DispatchService _dispatch;
    
    protected TaskBatch _batch;
    protected Timer _timer = new Timer();
    protected Hashtable _process = new Hashtable();
    
    class StartTask
        extends TimerTask
    {
        Task _task;
        public StartTask(Task task)
        {
            _task = task;
        }
        public void run()
        {
            try
            {
                _log.debug("starting " + _task);
                Process p = SystemUtil.command(_task.getFullCommand());
                _process.put(_task.getTaskId(), p);
            }
            catch (Exception ex)
            {
                _log.error(ex);
            }
        }
    }

    class KillTask
        extends TimerTask
    {
        Task _task;
        public KillTask(Task task)
        {
            _task = task;
        }
        public void run()
        {
            try
            {
                _log.debug("killing " + _task);
                Process p = (Process) _process.get(_task.getTaskId());
                if (p != null)
                {
                    p.destroy();
                }
                TaskFinishEvent tfe = new TaskFinishEvent(_task);
                _dispatch.dispatchEvent(
                    tfe, _task.getSource(), _forkName, _transportName);
            }
            catch (Exception ex)
            {
                _log.error(ex);
            }
        }
    }
    
    //=======================================================================
    //  Implementation of solar.service.Service interface.
    //=======================================================================

    public void init()
        throws ServiceException
    {
        _forkName = getName();
        Options config = getServiceManager().getServiceConfig(_forkName);
        _transportName = config.getProperty("transport", "transport");
    }
    
    public void initDependents()
        throws ServiceException
    {
        super.initDependents();     // must call at the top
        
        _dispatch = (DispatchService) 
            getDependentService("dispatch", "dispatch");
        _transportAddress = 
            _dispatch.getTransportAddress(_transportName);
        _rpc = (RpcService) getDependentService("rpc", "rpc");
    }

    public void handleEvent(Event evt)
    {
        // special trick if we use Transport instead of ObjectTransport
        if (TaskFinishEvent.isTaskFinishEvent(evt)) {
            TaskFinishEvent tfe = (evt instanceof TaskFinishEvent) ? 
                (TaskFinishEvent) evt : new TaskFinishEvent(evt);
                
            if (_batch != null) {
                _batch.taskFinished(tfe.getTask());
            }
        }
        else {
            throw new IllegalArgumentException(
                "Unexpected event received: " + evt);
        }
    }

    public Object handleRequest(Request request)
        throws Exception
    {
        if (request instanceof StartTaskRequest)
        {
            StartTaskRequest str = (StartTaskRequest) request;
            Task task = str.getTask();
            Id taskId = task.getTaskId();
            StartTask runtask = new StartTask(task);
            _timer.schedule(runtask, task.getStartTime()*1000);
            KillTask killtask = new KillTask(task);
            _timer.schedule(killtask, task.getEndTime()*1000);
        }
        else
        {
            throw new IllegalArgumentException(
                "Unexpected request received: " + request);
        }
        return null;
    }
    
    //=======================================================================
    // Implementation of ForkService interface.
    //=======================================================================
    
    public void start(Id id, Address dest, String cmd)
        throws ForkException
    {
        
    }

    public void destroy(Id id, Address dest)
        throws ForkException
    {
        
    }

    public void batchrun(String taskfile)
        throws ForkException
    {
        if (_batch != null)
        {
            throw new ForkException("Previous batch job not done yet.");
        }
        try
        {
            _batch = new TaskBatch(taskfile, this);
            _batch.run();
        }
        catch (Exception ex)
        {
            throw new ForkException(ex);
        }
    }

    public void startTask(Task task)
        throws Exception
    {
        task.setFinished(false);
        task.setSource(_transportAddress);
        StartTaskRequest sr = new StartTaskRequest(task);
        Address dest = task.getDestination();
        _rpc.invoke(dest, _forkName, sr, RPC_TIMEOUT);
    }

    public void batchFinished()
    {
        _batch = null;
        System.exit(0);
    }
}
