/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.fork;


import org.apache.log4j.Logger;
import java.io.FileReader;
import java.io.BufferedReader;
import java.util.List;
import java.util.Vector;
import java.util.Iterator;

import solar.api.ParsingException;
import solar.service.transport.SocketAddress;

/**
 *  A class contains batch jobs defined in a cfg file.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: TaskBatch.java,v 1.2 2003/10/08 22:53:22 glchen Exp $
 */
public class TaskBatch
{
    interface BatchListener
    {
        public void startTask(Task task)
            throws Exception;
            
        public void batchFinished();
    }
    
    class RepeatParam
    {
        String _param;
        int _start = 0;
        int _delta = 0;
        int _end = -1;
    }
    
    static Logger _log = Logger.getLogger(TaskBatch.class);
    
    int _repeat = 1;
//    List _paras = new Vector();
    RepeatParam _param = null;
    List _tasks = new Vector();
    BatchListener _listener;

    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public TaskBatch(String taskfile, BatchListener listener)
        throws Exception
    {
        parse(taskfile);
        _listener = listener;
    }

    //=======================================================================
    // Public interface methods.
    //=======================================================================

    public void run()
        throws Exception
    {
        while (_repeat != 0)
        {
            _repeat --;
            if (_param != null) {
                for (int i = _param._start; 
                    i <= _param._end; i += _param._delta) 
                {
                    runOneRound(_param._param, Integer.toString(i));
                    synchronized (_tasks) {
                        _tasks.wait();
                    }
                }
            }
            else {
                runOneRound(null, null);
                synchronized (_tasks) {
                    _tasks.wait();
                }
            }
        }
        _listener.batchFinished();
    }

    protected void runOneRound(String param, String value)
        throws Exception
    {
        Iterator itr = _tasks.iterator();
        while (itr.hasNext())
        {
            Task task = (Task) itr.next();
            if (param!=null && task.hasParameter(param))
            {
                task.setParameter(param, value);
            }
            _listener.startTask(task);
        }
    }

    public void taskFinished(Task task)
    {
        _log.debug("fininshed: " + task);
        boolean alldone = true;
        synchronized (_tasks)
        {
            Iterator itr = _tasks.iterator();
            while (itr.hasNext())
            {
                Task mytask = (Task) itr.next();
                if (mytask.getTaskId().equals(task.getTaskId()))
                {
                    mytask.setFinished(true);
                }
                else if (!mytask.isFinished())
                {
                    alldone = false;
                }
                
            }
            if (alldone)
            {
                _tasks.notifyAll();
            }
        }
    }
    
    //=======================================================================
    // Private helper methods.
    //=======================================================================
    
    private void parse(String taskfile)
        throws Exception
    {
        boolean repeatMode = true;
        Task task = null;
        
        BufferedReader in = new BufferedReader(new FileReader(taskfile));
        String line = null;
        while ( (line=in.readLine()) != null)
        {
            // process comments
            line = line.trim();
            if (line.startsWith("#")) continue;
            int idx = line.indexOf("#");
            if (idx != -1) line = line.substring(idx);
            if (line.length() == 0) continue;
            if (line.startsWith("[repeat]")) {
                repeatMode = true;
                if (task != null) {
                    _tasks.add(task);
                    task = null;
                }
                continue;
            }
            if (line.startsWith("[")) {
                repeatMode = false;
                if (task != null) {
                    _tasks.add(task);
                }
                task = new Task();
                idx = line.indexOf("]");
                String address = line.substring(1, idx);
                task.setDestination(new SocketAddress(address));
                continue;
            }
            if (repeatMode) {
                if (line.startsWith("-repeat")) {
                    idx = line.indexOf(" ");
                    line = line.substring(idx+1).trim();
                    _repeat = Integer.parseInt(line);
                }
                else if (line.startsWith("-")) {
                    throw new ParsingException("Syntax error " + line);
                }
                else {
                    String[] f = line.split(" ");
                    RepeatParam rp = new RepeatParam();
                    rp._param = f[0];
                    rp._start = Integer.parseInt(f[1]);
                    rp._delta = Integer.parseInt(f[2]);
                    rp._end = Integer.parseInt(f[3]);
//                    _paras.add(rp);
                    _param = rp;
                }
            }
            else {
                if (line.startsWith("-cmd")) {
                    idx = line.indexOf(" ");
                    task.setCommand(line.substring(idx+1));
                }
                else if (line.startsWith("-start")) {
                    idx = line.indexOf(" ");
                    line = line.substring(idx+1).trim();
                    task.setStartTime(Integer.parseInt(line));
                }
                else if (line.startsWith("-end")) {
                    idx = line.indexOf(" ");
                    line = line.substring(idx+1).trim();
                    task.setEndTime(Integer.parseInt(line));
                }
                else if (line.startsWith("-")) {
                    throw new ParsingException("Syntax error " + line);
                }
                else {
                    idx = line.indexOf(" ");
                    task.setParameter(
                        line.substring(0, idx), line.substring(idx+1));
                }
            }
        }
        if (task != null) {
            _tasks.add(task);
        }
    }
}
