/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.multicast;


import java.io.Serializable;
import java.util.Hashtable;
import rice.scribe.Scribe;
import rice.scribe.IScribe;
import rice.scribe.IScribeApp;
import rice.scribe.messaging.ScribeMessage;
import rice.pastry.NodeId;
import rice.pastry.NodeHandle;
import rice.pastry.security.Credentials;

import solar.api.Id;
import solar.api.Event;
import solar.service.BaseService;
import solar.service.ServiceException;
import solar.service.transport.Transport;
import solar.service.transport.TransportService;
import solar.service.transport.PastryTransport;


/**
 *  This multicast service adapts the Scribe substrate based on Pastry.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: ScribeMulticastService.java,v 1.3 2003/09/25 21:37:39 glchen Exp $
 */
public class ScribeMulticastService 
    extends BaseService
    implements MulticastService
{
    protected Scribe _scribe;
    protected Credentials _cred;
    protected PastryTransport _pastry;
    protected Hashtable _members;

    
    //=======================================================================
    //  Implementation of solar.service.Service interface.
    //=======================================================================
    
    public void initDependents()
        throws ServiceException
    {
        super.initDependents();     // must call at the top
        
        TransportService transport = (TransportService) 
            getDependentService("transport", "transport");
        Transport tp = transport.getTransport();
        if (!(tp instanceof PastryTransport))
        {
            throw new ServiceException(
                "Requires PastryTransportService for Scribe mutlicast.");
        }
        _pastry = (PastryTransport) tp;
        _cred = _pastry.getCredentials();
        _scribe = new Scribe(
            _pastry.getPastryNode(), _pastry.getCredentials());
        _members = new Hashtable();
    }


    //=======================================================================
    //  Implementation of solar.service.multicast.MulticastService interface.
    //=======================================================================

    public void create(Id groupId)
        throws MulticastException
    {
        _scribe.create(_scribe.generateTopicId(groupId.toString()), _cred);
    }

    public void destroy(Id groupId)
        throws MulticastException
    {
        // not possible in Scribe?
    }

    public void join(Id groupId, MulticastMember mcmember)
        throws MulticastException
    {
        ScribeMember sm = new ScribeMember(mcmember);
        _members.put(mcmember, sm);
        _scribe.registerApp(sm);
        _scribe.join(_scribe.generateTopicId(groupId.toString()), sm, _cred);
    }

    public void leave(Id groupId, MulticastMember mcmember)
        throws MulticastException
    {
        ScribeMember sm = (ScribeMember) _members.get(mcmember);
        if (sm != null)
            _scribe.leave(_scribe.generateTopicId(groupId.toString()), sm, _cred);
    }

    public void multicast(Id groupId, Event evt)
        throws MulticastException
    {
        _scribe.multicast(_scribe.generateTopicId(groupId.toString()), evt, _cred);
    }


    //=======================================================================
    //  Helper class.
    //=======================================================================

    class ScribeMember
        implements IScribeApp
    {
        MulticastMember _mcmember;
        
        //===================================================================
        //  Constructors.
        //===================================================================

        public ScribeMember(MulticastMember mcmember)
        {
            _mcmember = mcmember;
        }

        //===================================================================
        //  Implementation of rice.scribe.IScribeApp interface.
        //===================================================================

        /**
         * Invoked when the underlying Scribe substrate is ready. 
         * The Scribe substrate becomes ready as soon as the local Pastry node
         * on which it lies is ready. In order to get this upcall it is 
         * necessary that the IScribeApp registers itself to the Scribe 
         * substrate using the registerApp() method in IScribe interface.
         */
        public void scribeIsReady()
        {
        
        }

        /**
         * up-call invoked by scribe when a publish message is 'delivered'.
         */    
        public void receiveMessage(ScribeMessage msg)
        {
            _mcmember.handleMulticastEvent((Event)msg.getData());
        }

        /**
         * Called by Scribe before the node forwards a message to
         * its children in the multicast tree.
         */
        public void forwardHandler(ScribeMessage msg)
        {
            
        }

         /**
         * Invoked by Scribe after a child is added to or removed from one of
         * the node's children tables.
         */
        public void subscribeHandler(NodeId topicId, 
            NodeHandle child, boolean wasAdded, Serializable obj )
        {
            
        }
    
        /**
         * Invoked by Scribe just before the "repair" SUBSCRIBE message is sent
         * when a node suspects its parent is faulty.
         */
        public void faultHandler(ScribeMessage msg, NodeHandle faultyParent)
        {
            
        }

        /**
         * Invoked by underlying scribe when it receives a anycast Message, 
         * to check if there is a local application which can take care of 
         * this anycast message.
         * If application can service the request of anycast message, the
         * anycast message is not routed furthur(for DFS), else the anycast
         * message will be furthur routed following DFS.
         */
        public boolean anycastHandler(ScribeMessage msg)
        {
            return true;
        }

        /**
         * Upcall made by scribe to registered applications to
         * inform them that local node is now the root for some
         * topic, and hence can take appropriate action.
         */
        public void isNewRoot(NodeId topicId)
        {
            
        }

        /**
         * Upcall made by scribe to the registered applications to
         * inform them that a new parent is found for local node
         * in the topic tree.
         */
        public void newParent(NodeId topicId, 
            NodeHandle newParent, Serializable data)
        {
            
        }

    }

}
