/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.pack;


import serp.util.Options;
import org.apache.log4j.Logger;
import java.util.Set;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Timer;
import java.util.TimerTask;

import solar.api.Id;
import solar.api.Guid;
import solar.api.Event;
import solar.api.Address;
import solar.api.Request;
import solar.module.protocol.Protocol;
import solar.service.BaseService;
import solar.service.ServiceException;
import solar.service.dispatch.DispatchService;
import solar.service.pack.proto.BufferedEvent;
import solar.service.pack.proto.InputProtocol;
import solar.service.pack.proto.OutputProtocol;
import solar.service.pack.proto.ForwardProtocol;
import solar.service.pack.comm.AckEvent;
import solar.service.pack.comm.PackEvent;
import solar.service.pack.comm.SubscribeEvent;
import solar.service.pack.comm.SubscribeRequest;
import solar.service.pack.comm.ScribeRootRequest;
import solar.service.pack.comm.AttachPublisherRequest;
import solar.service.pack.comm.AttachSubscriberRequest;
import solar.service.pack.comm.DetachPublisherRequest;
import solar.service.pack.comm.DetachSubscriberRequest;
import solar.service.pubsub.Publisher;
import solar.service.pubsub.Subscriber;
import solar.service.rpc.RpcService;
import solar.service.rpc.RpcException;
import solar.service.rpc.RpcConstants;
import solar.service.transport.PastryTransport;
import solar.service.transport.TransportService;


/**
 *  The PACK service that runs at the server side and is built on top
 *  of Scribe.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: ScribePackService.java,v 1.35 2003/10/13 19:34:25 glchen Exp $
 */
public class ScribePackService 
    extends BaseService
    implements PackService, RpcConstants, PackConstants
{
    /** Logging facility. */
    protected static Logger _log = 
        Logger.getLogger(ScribePackService.class);

    protected String _packName;
    protected String _coreTransportName;
    protected String _edgeTransportName;
    protected Address _coreTransportAddress;

    protected RpcService _coreRpc;
    protected DispatchService _dispatch;
    protected TransportService _coreTransport;

    protected ScribeManager _scribeman;
    protected Hashtable  _scribeRoots = new Hashtable(); // (Id, rootAddress)
    protected Hashtable _scribeProxies = new Hashtable(); // (Id, proxyAddress)

    // input protocol runs for subscribers in same address space
    protected InputProtocol _inproto;
    // output protocol runs for directly connected clients
    protected OutputProtocol _outproto;
    // forward protocol runs between servers on Scribe tree
    protected ForwardProtocol _fwdproto;
    // root protocol runs between proxy and Scribe root
    protected OutputProtocol _pubrootproto;
    protected OutputProtocol _subrootproto;

    protected Hashtable _fakeids = new Hashtable();
    protected EndPointManager _epman = new EndPointManager();

    protected Timer _timer = new Timer(true);
    protected Hashtable _subtasks = new Hashtable(); // (pubId, (subId, SubscribeTask)
    protected Hashtable _roottasks = new Hashtable(); // (Id, RootTask)

    //=======================================================================
    // Implementation of solar.service.Service interface.
    //=======================================================================

    public void init()
        throws ServiceException
    {
        _packName = getName();
        Options config = getServiceManager().getServiceConfig(_packName);
        _coreTransportName = config.getProperty("core_transport", "transport");
        _edgeTransportName = config.getProperty("edge_transport", "transport");

        int bt = 
            config.getIntProperty("buffer_limit", PACK_BUFFER_LIMIT);
        int mt = 
            config.getIntProperty("memory_limit", PACK_MEMORY_LIMIT);
        int st = 
            config.getIntProperty("suspend_limit", PACK_SUSPEND_LIMIT);
            
        _inproto = new InputProtocol(bt, mt, st);
        _outproto = new OutputProtocol(bt, mt, st);
        _fwdproto = new ForwardProtocol(bt, mt, st);
        _pubrootproto = new OutputProtocol(bt, mt, st);
        _subrootproto = new OutputProtocol(bt, mt, st);

        new Thread(new Handler(), "pack-input").start();
        new Thread(new Sender(), "pack-output").start();
        new Thread(new Forwarder(_fwdproto), "pack-forwarder").start();
        new Thread(new Forwarder(_pubrootproto), "pack-pubroot").start();
        new Thread(new Forwarder(_subrootproto), "pack-subroot").start();
        
        initDependents();
    }

    public void initDependents()
        throws ServiceException
    {
        super.initDependents();     // must call at the top

        // retrieve needed services
        _dispatch = (DispatchService) 
            getDependentService("dispatch", "dispatch");
        _coreRpc = (RpcService)
            getDependentService("core_rpc", "core_rpc");
        _coreTransport = (TransportService)
            getDependentService("core_transport", "core_transport");
        _coreTransportAddress = _coreTransport.getLocalAddress();

        // get scribe manager
        //TransportService pastryTransport = (TransportService) 
        //    getDependentService("pastry_transport", "pastry_transport");
        PastryTransport pastry = (PastryTransport) _coreTransport.getTransport();
        _scribeman = new ScribeManager(
            pastry.getPastryNode(), pastry.getCredentials());
    }
    
    public Object handleRequest(Request request)
        throws Exception
    {
        Object ret = null;
        if (request instanceof AttachSubscriberRequest) {
            AttachSubscriberRequest asr = (AttachSubscriberRequest) request;
            Id session = makeSession();
            Id subId = asr.getSubId();
            doAttachSubscriber(subId, session, asr.getSourceAddress(), null);
            ret = session;
        }
        else if (request instanceof AttachPublisherRequest) {
            AttachPublisherRequest apr = (AttachPublisherRequest) request;
            Id session = makeSession();
            Id pubId = apr.getPubId();
            doAttachPublisher(pubId, session, apr.getSourceAddress(), null);
            ret = session;
        }
        else if (request instanceof DetachSubscriberRequest) {
            DetachSubscriberRequest dsr = (DetachSubscriberRequest) request;
            Id subId = dsr.getSubId();
            doDetachSubscriber(subId);
        }
        else if (request instanceof DetachPublisherRequest) {
            DetachPublisherRequest dpr = (DetachPublisherRequest) request;
            Id pubId = dpr.getPubId();
            doDetachPublisher(pubId);
        }
        else if (request instanceof SubscribeRequest) {
            SubscribeRequest sr = (SubscribeRequest) request;
            Id pubId = sr.getPubId();
            Id subId = sr.getSubId();
            Object policy = sr.getPolicy();
            doSubscribe(pubId, subId, policy);
        }
        else if (request instanceof ScribeRootRequest) {
            ScribeRootRequest srr = (ScribeRootRequest) request;
            Id id = srr.getId();
            _log.debug("got root request for " + id);
            addScribeProxy(id, srr.getSourceAddress());
            // return local address telling proxy I am the root
            ret = _coreTransportAddress;
        }
        else {
            throw new RpcException(
                "Unexpected pack request: " + request);
        }
        return ret;
    }

    public void handleEvent(Event evt)
        throws Exception
    {
        // special trick if we use Transport instead of ObjectTransport
        if (PackEvent.isPackEvent(evt)) {
            PackEvent pe = (evt instanceof PackEvent) ? 
                (PackEvent) evt : new PackEvent(evt);
            handlePackEvent(pe);
        }
        else if (SubscribeEvent.isSubscribeEvent(evt)) {
            SubscribeEvent se = (evt instanceof SubscribeEvent) ? 
                (SubscribeEvent) evt : new SubscribeEvent(evt);
            handleSubscribeEvent(se);
        }
        else {
            throw new IllegalArgumentException(
                "Unexpected pack event: " + evt);
        }
    }

    protected void handleSubscribeEvent(SubscribeEvent evt)
        throws Exception
    {
        _log.debug("handle subscribe event");
        Id pubId = evt.getPubId();
        Id subId = evt.getSubId();
        Object policy = evt.getPolicy();
        Address lasthop = evt.getLastHop();
        try {
            // join multicast if this node is the subscriber root
            if (_scribeman.isRoot(subId)) {
                _log.debug("subscription reached sub root");
                // make local scribe join the topic's multicast tree
                _scribeman.join(pubId);
                // set up the forwarding path to subscriber proxy
                Id fakeId = makeFakeId(pubId, subId);
                addFakeId(pubId, fakeId);
                _log.debug("subrootproto: add " + fakeId + " for " + subId);
                _subrootproto.addPublisher(fakeId);
                _subrootproto.addSubscriber(fakeId, subId, policy, lasthop);
                _subrootproto.resumeSubscriber(subId);
            }
            // check if this is the final stop for the subscribe event
            if (_epman.contains(pubId) && !evt.isForwarding()) {
                // this is the publisher proxy
                _log.debug("subscription reached pub proxy");
                // set up forwarding path to publisher root
                _log.debug("pubrootproto: add " + pubId + " from " + lasthop);
                _pubrootproto.addPublisher(pubId);
                _pubrootproto.addSubscriber(
                    pubId, subId, policy, getScribeRoot(pubId));
                Address addr = _epman.getAddress(pubId);
                if (addr != null) {
                    // relay subscription to pub client
                    sendToClient(addr, evt);
                    _log.debug("subscription sent to pub client.");
                }
                return;     // stop popup the events
            }
            // create new subscribe event to popup the multicast tree
            if (_scribeman.isRoot(pubId)) {
                _log.debug("subscription reached pub root");
                // create the Scribe topic
                _scribeman.create(pubId);
                // allocate forwarding buffer to children
                if (!isLocal(lasthop)) {
                    _log.debug("fwdproto: add " + pubId);
                    _fwdproto.addPublisher(pubId);
                    _fwdproto.addSubscriber(pubId, subId, policy, lasthop);
                }
                // create new subscribe event indicating forwarding is ended
                evt = new SubscribeEvent(
                    pubId, subId, policy, _coreTransportAddress, false);
                // relay subscription to pub proxy
                if (!sendToProxy(pubId, evt)) {
                    handleSubscribeEvent(evt);
                }
                else {
                    _log.debug("subscription sent to pub proxy: " + evt);
                }
            }
            else {
                // allocate forwarding buffer to children
                _log.debug("fwdproto: add " + pubId);
                _fwdproto.addPublisher(pubId);
                _fwdproto.addSubscriber(pubId, subId, policy, lasthop);
                evt = new SubscribeEvent(
                    pubId, subId, policy, _coreTransportAddress, true);
                // relay subscription to pub parent
                sendToParent(pubId, evt);
                _log.debug("subscription sent to pub parent");
            }
        }
        catch (Exception ex) { _log.warn(ex); }
    }

    protected void handlePackEvent(PackEvent evt)
        throws Exception
    {
        Id pubId = evt.getPubId();
        Id subId = evt.getSubId();
        if (subId == null) {
            // fresh events from clients, set forwarding to be false
            // forwarding flag will be turned on at the publisher root
            evt = new PackEvent(pubId, pubId, evt.getEvent(), false);
            if (_pubrootproto.hasPublisher(pubId)) {
                _pubrootproto.putEvent(evt, null);
            }
            else {
                _log.error(
                    "Publisher " + pubId + " not found in root protocol.");
            }
        }
        else {
            doPublish(pubId, evt);
        }
    }

    //=======================================================================
    // Implementation of PackService interface.
    //=======================================================================
    
    public void attachPublisher(Id pubId, Publisher publisher)
        throws PackException
    {
        _epman.assertNotAttached(pubId);
        doAttachPublisher(pubId, makeSession(), null, publisher);
    }

    public void detachPublisher(Id pubId)
        throws PackException
    {
        try {
            doDetachPublisher(pubId);
        }
        catch (Exception ex) {
            throw new PackException(ex);
        }
    }

    public void attachSubscriber(Id subId, Subscriber subscriber)
        throws PackException
    {
        _epman.assertNotAttached(subId);
        doAttachSubscriber(subId, makeSession(), null, subscriber);
    }

    public void detachSubscriber(Id subId)
        throws PackException
    {
        try {
            doDetachSubscriber(subId);
        }
        catch (Exception ex) {
            throw new PackException(ex);
        }
    }
    
    public void subscribe(Id pubId, Id subId, Object policy)
        throws PackException
    {
        _epman.assertAttached(subId);
        doSubscribe(pubId, subId, policy);
    }

    public void publish(Id pubId, Event evt)
        throws PackException
    {
        _epman.assertAttached(pubId);
        doPublish(pubId, new PackEvent(pubId, pubId, evt, false));
    }

    public Guid makeGuid(Id id)
    {
        return new Guid(_coreTransportAddress, id);
    }

    public static Id makeFakeId(Id pubId, Id subId)
    {
        return new Id(pubId.toString() + "-" + subId.toString());
    }

    //=======================================================================
    // Private helper methods.
    //=======================================================================

    protected void doAttachPublisher(Id pubId, Id session, 
        Address address, Publisher publisher)
    {
        RootTask roottask = new RootTask(pubId);
        // execute once and schedule periodic task
        roottask.run();
        _timer.scheduleAtFixedRate(roottask, 30000, 30000);
        _roottasks.put(pubId, roottask);
        // record this publisher as an end point
        _epman.attach(pubId, session, address, publisher);
    }

    protected void doDetachPublisher(Id pubId)
        throws Exception
    {
        // remove pubId from pubrootproto and epman
        _epman.detach(pubId);
        _pubrootproto.removePublisher(pubId);            
        // remove the root task for pubId
        removeRootTask(pubId);
        // tell the root to cancel its proxy
        Address root = getScribeRoot(pubId);
        if (root == null ||isLocal(root)) {
            removeScribeProxy(pubId);
        }
        else {
            callRpc(root, new DetachPublisherRequest(pubId));
        }
    }
    
    protected void doAttachSubscriber(Id subId, Id session, 
        Address address, Subscriber subscriber)
    {
        RootTask roottask = new RootTask(subId);
        // execute once and schedule periodic task
        roottask.run();
        _timer.scheduleAtFixedRate(roottask, 30000, 30000);
        _roottasks.put(subId, roottask);
        // record this subscriber as an end point
        _epman.attach(subId, session, address, subscriber);
    }

    protected void doDetachSubscriber(Id subId)
        throws Exception
    {
        // remove subId from epman
        _epman.detach(subId);
        // remove the root task and all its subscribe task
        removeRootTask(subId);
        removeSubscribeTask(subId);
        // tell the root to suspend subId from its subrootproto
        Address root = getScribeRoot(subId);
        if (root == null || isLocal(root)) {
            _subrootproto.suspendSubscriber(subId);
            removeScribeProxy(subId);
        }
        else {
            callRpc(root, new DetachSubscriberRequest(subId));
        }
        removeScribeRoot(subId);
    }
    
    protected void doSubscribe(Id pubId, Id subId, Object policy)
        throws PackException
    {
        // allocate outbound buffers or inbound buffers
        Address client = _epman.getAddress(subId);
        Id fakeId = makeFakeId(pubId, subId);
        if (client != null) { // subscribe is directly connected client
            _log.debug("outproto: add " + fakeId);
            _outproto.addPublisher(fakeId);
            _outproto.addSubscriber(fakeId, subId, policy, client);
        }
        else { // subscribe in same address space
            _log.debug("inproto: add " + fakeId);
            _inproto.addPublisher(fakeId);
            _inproto.addSubscriber(fakeId, subId, policy);
        }
        // popup the subscribe event to subscriber's root and publisher's proxy
        SubscribeEvent evt = new SubscribeEvent(
            pubId, subId, policy, _coreTransportAddress, true);
        SubscribeTask subtask = new SubscribeTask(subId, evt);
        _timer.scheduleAtFixedRate(subtask, 0, 5000);
        addSubscribeTask(pubId, subId, subtask);
    }

    protected void doPublish(Id pubId, Event evt)
        throws PackException
    {
        PackEvent pe = (PackEvent) evt;
        Id subId = pe.getSubId();
//        _log.debug("received event for " + pubId);
        if (_inproto.hasPublisher(pubId)) {
            _inproto.putEvent(evt, null);
        }
        if (_outproto.hasPublisher(pubId)) {
//            _log.debug("outproto: got " + pubId);
            _outproto.putEvent(evt, null);
        }
        if (!pe.isForwarding() && _scribeman.isRoot(pubId)) {
            // start forwarding this event
            pe = new PackEvent(pubId, subId, pe.getEvent(), true);
        }
        if (pe.isForwarding() && _fwdproto.hasPublisher(pubId)) {
//            _log.debug("fwdproto: got " + pubId);
            _fwdproto.putEvent(evt, null);
        }
        Set ids = getFakeIds(pubId);
        if (ids == null) return;
        Iterator itr = ids.iterator();
        while (itr.hasNext()) {
            Id fakeId = (Id) itr.next();
//            evt = new PackEvent(fakeId, subId, pe.getEvent(), false);
            evt = new PackEvent(fakeId, fakeId, pe.getEvent(), false);
            if (_subrootproto.hasPublisher(fakeId)) {
                _log.debug("subroot: got " + fakeId);
                _subrootproto.putEvent(evt, null);
            }
        }
    }

    //=======================================================================
    // Inner helper classes.
    //=======================================================================

    /** Periodically update the (pubId, rootAddress) mapping. */
    class RootTask extends TimerTask
    {
        Id _id;
        public RootTask(Id id) {
            _id = id;
        }
        public void run() {
            try {
                // request to find the root of this pubId
                Request request = new ScribeRootRequest(_id);
                request.setSourceAddress(_coreTransportAddress);
                Address dest = _scribeman.getNodeAddress(_id);
                Address root = (Address) 
                    _coreRpc.invoke(dest, _packName, request, RPC_TIMEOUT);
                addScribeRoot(_id, root);
            }
            catch (Exception ex) { _log.warn(ex); }
        }
    }
    
    /** Periodically refresh the subscription in the Scribe network. */
    class SubscribeTask extends TimerTask
    {
        Id _subId;
        SubscribeEvent _subevt;
        public SubscribeTask(Id subId, SubscribeEvent evt) {
            _subId = subId;
            _subevt = evt;
        }
        public void run() {
            try {
                if (!sendToRoot(_subId, _subevt)) {
                    handleSubscribeEvent(_subevt);
                }
                else {
                    _log.debug("sent subscription to root");
                }
            }
            catch (Exception ex) { _log.warn(ex); }
        }
    }

    /** Responsible to delivery events to children in multicast tree. */
    class Forwarder implements Runnable
    {
        Protocol _protocol;
        public Forwarder(Protocol protocol) {
            _protocol = protocol;
        }
        public void run() {
            while (true) {
                try {
                    BufferedEvent evt = (BufferedEvent) _protocol.getEvent();
                    Address dest = evt.getDestination();
                    if (dest == null) { // multicast -- not used anymore
                        Id pubId = evt.getPubId();
                        Iterator itr = _scribeman.getChildren(pubId).iterator();
                        while (itr.hasNext()) {
                            Address next = (Address) itr.next();
                            _dispatch.dispatchEvent(
                                evt.getEvent(),
                                next, _packName, _coreTransportName);
                        }
                    }
                    else {  // unicast
                        if (isLocal(dest)) {
                            handlePackEvent((PackEvent) evt.getEvent());
                        }
                        else {
                            _dispatch.dispatchEvent(
                                evt.getEvent(),
                                dest, _packName, _coreTransportName);
                        }
                    }
                }
                catch (Exception ex) {
                    _log.error(ex);
                    break;
                }
            }
        }
    }

    /** Responsible to delivery events to attached clients only. */
    class Sender implements Runnable
    {
        public void run() {
            while (true) {
                try {
                    BufferedEvent evt = (BufferedEvent) _outproto.getEvent();
                    Address dest = evt.getDestination();
                    _dispatch.dispatchEvent(
                        evt.getEvent(),
                        dest, _packName, _edgeTransportName);
                }
                catch (Exception ex) {
                    _log.error(ex);
                    break;
                }
            }
        }
    }
    
    //TODO: use thread pool to call subscriber's event handler
    /** Responsible to delivery events to local subscribers only. */
    class Handler implements Runnable
    {
        public void run() {
            while (true) {
                try {
                    BufferedEvent evt = (BufferedEvent) _inproto.getEvent();
                    Id subId = evt.getSubId();
                    Subscriber subscriber = 
                        (Subscriber) _epman.getEndPoint(subId);
                    if (subscriber != null) {
                        PackEvent pe = (PackEvent) evt.getEvent();
                        subscriber.handleEvent(pe.getEvent());
                    }
                    else {
                        _log.error("Received event for non-exist subscriber " + subId);
                    }
                }
                catch (Exception ex) {
                    _log.error(ex);
                    break;
                }
            }
        }
    }

    //=======================================================================
    // Private help methods.
    //=======================================================================

    private Object callRpc(Address dest, Request request)
        throws Exception
    {
        return _coreRpc.invoke(dest, _packName, request, RPC_TIMEOUT);
    }
    
    /** Is the specified address actually same as local address. */
    private boolean isLocal(Address dest)
    {
        return _coreTransportAddress.equals(dest);
    }

    /** Send an event to client using edge_transport. */
    private void sendToClient(Address dest, Event evt)
        throws Exception
    {
        _dispatch.dispatchEvent(evt, dest, _packName, _edgeTransportName);
    }

    /** Send an event to a topic's parent, return false if parent is null. */
    private boolean sendToParent(Id pubId, Event evt)
        throws Exception
    {
        Address parent = _scribeman.getParent(pubId);
        if (parent == null)
            return false;

        _dispatch.dispatchEvent(
            evt, parent, _packName, _coreTransportName);
        return true;
    }

    /** Send an event to a topic's root, return false if root is local. */
    private boolean sendToRoot(Id id, Event evt)
        throws Exception
    {
        Address root = getScribeRoot(id);
        if (root == null)
            throw new NullPointerException("Root not found for " + id);
            
        if (isLocal(root)) return false;

        _dispatch.dispatchEvent(
            evt, root, _packName, _coreTransportName);
        return true;
    }

    /** Send an event to a topic's proxy, return false if proxy is local. */
    private boolean sendToProxy(Id id, Event evt)
        throws Exception
    {
        Address proxy = getScribeProxy(id);
        if (proxy == null) 
            throw new NullPointerException("Proxy not found for " + id);

        if (isLocal(proxy)) return false;

        _dispatch.dispatchEvent(evt, proxy, _packName, _coreTransportName);
        return true;
    }

    /** Return a new session ID. */
    private Id makeSession()
    {
        String stamp = Long.toString(System.currentTimeMillis());
        return Id.makeId("session-" + stamp + "-");
    }

    private void addRootTask(Id id, RootTask roottask)
    {
        synchronized (_roottasks) {
            _roottasks.put(id, roottask);
        }
    }

    private void removeRootTask(Id id)
    {
        synchronized (_roottasks) {
            RootTask roottask = (RootTask) _roottasks.remove(id);
            if (roottask != null) {
                roottask.cancel();
            }
        }
    }

    private void addSubscribeTask(Id pubId, Id subId, SubscribeTask subtask)
    {
        synchronized (_subtasks) {
            Hashtable t = (Hashtable) _subtasks.get(pubId);
            if (t == null) {
                t = new Hashtable();
                _subtasks.put(pubId, t);
            }
            t.put(subId, subtask);
        }
    }

    private void removeSubscribeTask(Id pubId, Id subId)
    {
        synchronized (_subtasks) {
            Hashtable t = (Hashtable) _subtasks.get(pubId);
            if (t != null) {
                SubscribeTask subtask = (SubscribeTask) t.remove(subId);
                if (subtask != null) {
                    subtask.cancel();
                }
            }        
        }
    }

    private void removeSubscribeTask(Id subId)
    {
        synchronized (_subtasks) {
            Iterator itr = _subtasks.values().iterator();
            while (itr.hasNext()) {
                Hashtable t = (Hashtable) itr.next();
                SubscribeTask subtask = (SubscribeTask) t.remove(subId);
                if (subtask != null) {
                    subtask.cancel();
                }
            }
        }
    }

    private void addScribeRoot(Id id, Address root)
    {
        synchronized (_scribeRoots) {
            _scribeRoots.put(id, root);
        }
    }

    private void removeScribeRoot(Id id)
    {
        synchronized (_scribeRoots) {
            _scribeRoots.remove(id);
        }
    }

    private Address getScribeRoot(Id id)
    {
        synchronized (_scribeRoots) {
            return (Address) _scribeRoots.get(id);
        }
    }

    private void addScribeProxy(Id id, Address proxy)
    {
        synchronized (_scribeProxies) {
            _scribeProxies.put(id, proxy);
        }
    }

    private void removeScribeProxy(Id id)
    {
        synchronized (_scribeProxies) {
            _scribeProxies.remove(id);
        }
    }

    private Address getScribeProxy(Id id)
    {
        synchronized (_scribeProxies) {
            return (Address) _scribeProxies.get(id);
        }
    }

    private void addFakeId(Id pubId, Id fakeId)
    {
        synchronized (_fakeids) {
            Set ids = (Set) _fakeids.get(pubId);
            if (ids == null) {
                ids = new HashSet();
                _fakeids.put(pubId, ids);
            }
            ids.add(fakeId);
        }
    }

    private Set getFakeIds(Id pubId)
    {
        synchronized (_fakeids) {
            return (Set) _fakeids.get(pubId);
        }        
    }
}
