/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.pack.lang;


import org.xml.sax.*;
import org.xml.sax.helpers.*;
import java.io.File;
import java.io.Reader;
import java.io.StringReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.FileInputStream;
import java.util.List;
import java.util.Vector;
import javax.xml.parsers.*;

import solar.api.ParsingException;
import solar.service.pack.PackPolicy;


/**
 *  An XML parser for PACK policies.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: XmlPolicyParser.java,v 1.1 2003/09/25 21:54:00 glchen Exp $
 */
public class XmlPolicyParser 
	extends DefaultHandler
    implements PolicyParser
{
	public static final String PACK         = "pack";
    public static final String GROUP        = "group";
    public static final String KEYS         = "keys";
	public static final String FILTER       = "filter";
	public static final String NAME         = "name";
	public static final String VALUE        = "value";
	public static final String ARGUMENT     = "argument";

	private static final SAXParserFactory _factory = 
		SAXParserFactory.newInstance();
    
    protected PackPolicy _policy;
    protected List _filters;        // previous parsed filters
    protected FilterConfig _filter; // current filter def
    
    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public XmlPolicyParser()
    {
    }


    //=======================================================================
    // Implementation of solar.service.pack.PolicyParser interface.
    //=======================================================================

    public PackPolicy parse(File file)
        throws ParsingException
    {
		try
		{
			FileInputStream in = new FileInputStream(file);
			PackPolicy ret = parse(in);
			in.close();
			return ret;
		}
		catch (IOException ioe)
		{
			throw new ParsingException(ioe);
		}
    }

    public PackPolicy parse(InputStream in)
        throws ParsingException
    {
		return parse(new InputSource (in));
    }

    public PackPolicy parse(Reader reader)
        throws ParsingException
    {
		return parse(new InputSource(reader));
    }

    public PackPolicy parse(String xml)
        throws ParsingException
    {
		StringReader reader = new StringReader(xml);
		return parse(new InputSource(reader));
    }


    //=======================================================================
    // Public interface methods.
    //=======================================================================

	public static PackPolicy parse(InputSource in)
		throws ParsingException
	{
		try
		{
			SAXParser sax = _factory.newSAXParser();
			XmlPolicyParser parser = new XmlPolicyParser();
			sax.parse(in, parser);
			return parser.getPackPolicy();
		}
        catch (SAXException saxe)
        {
			throw new ParsingException(saxe);
        }
		catch (ParserConfigurationException pce)
		{
			throw new ParsingException(pce);
		}
		catch (IOException ioe)
		{
			throw new ParsingException(ioe);
		}
	}


    //=======================================================================
    // Parsing methods.
    //=======================================================================

    /** Return recent parsed PACK policy. */
    public PackPolicy getPackPolicy()
    {
        return _policy;
    }

	public void startElement (String uri, String localName, 
        String name, Attributes attrs)
		throws SAXException
	{
		switch (name.charAt (0))
		{
		case 'p':
            assertEquals(name, PACK);
            startPolicy(attrs);
			break;
		case 'g':
            assertEquals(name, GROUP);
            startGroup(attrs);
			break;
		case 'f':
            assertEquals(name, FILTER);
            startFilter(attrs);
			break;
		case 'a':
            assertEquals(name, ARGUMENT);
            startArgument(attrs);
			break;
		default:
			assertEquals(name, "");	// will throw proper exception
		}
    }

	public void endElement(String uri, String localName, String name)
		throws SAXException
	{
		switch (name.charAt (0))
		{
		case 'f':
            assertEquals(name, FILTER);
            endFilter();
			break;
		case 'p':
            assertEquals(name, PACK);
            endPolicy();
			break;
		}
	}

    private void startPolicy(Attributes attrs)
    {
        _policy = new PackPolicy();
        _filters = new Vector();
    }

    private void startGroup(Attributes attrs)
    {
        String[] group = attrs.getValue(KEYS).split(" ");
        _policy.setGroupKeys(group);
    }

    private void startFilter(Attributes attrs)
    {
        _filter = new FilterConfig();
        _filter.setName(attrs.getValue(NAME));
    }

    private void endFilter()
    {
        _filters.add(_filter);
    }

    private void endPolicy()
		throws SAXException
    {
        try
        {
            _policy.setFilters(_filters);
        }
        catch (Exception ex)
        {
            throw new SAXException(ex);
        }
    }

    private void startArgument(Attributes attrs)
    {
        String name = attrs.getValue(NAME);
        String value = attrs.getValue(VALUE);
        _filter.addArgument(name, value);
    }
    
	private void assertEquals (String s1, String s2)
		throws SAXException
	{
		if (!(s1.equals (s2)))
        {
			throw new SAXException("Illegal tag: " + s1);
        }
	}


    //=======================================================================
    // Test.
    //=======================================================================

    public static void main(String[] args)
        throws Exception
    {
        PolicyParser p = new XmlPolicyParser();
        p.parse(new File(args[0]));
    }
}
