/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.pack.proto;


import cern.jet.random.Uniform;
import cern.jet.random.engine.MersenneTwister;
import java.util.List;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Iterator;

import solar.api.Id;
import solar.api.Event;
import solar.api.Address;
import solar.module.protocol.Buffer;
import solar.module.protocol.BufferFullException;


/**
 *  Manages a group of subscriptions, such as those coming from same link.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: GroupBufferHandle.java,v 1.9 2003/10/13 19:34:26 glchen Exp $
 */
class GroupBufferHandle 
    implements BufferHandle
{
    protected Uniform _unigen;
    
    protected int _size;
    protected Id _pubId;
    protected Object _subkey;           // link address
    protected Hashtable _policy;        // (subId, policy)
    protected Buffer _sharedBuffer;     // shared buffer for all group memebers

    // we force the suspend request is applied to all subscribers
    // in this handle, such as outproto at client and subrootproto
    protected boolean _suspended;
    protected int _suspendLimit;
    protected Object _plock = new Object();
    protected Hashtable _privateBuffers;    // (subId, SingleBufferHandle)
    
    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public GroupBufferHandle(Id pubId, Object subkey, int suspendLimit)
    {
        _pubId = pubId;
        _subkey = subkey;
        _policy = new Hashtable();
        _sharedBuffer = new Buffer();
        _privateBuffers = new Hashtable();
        _suspendLimit = suspendLimit;
        _unigen = new Uniform(new MersenneTwister());
    }

    //=======================================================================
    // Implementation of BufferHandle interface.
    //=======================================================================

    public int size()
    {
        synchronized (_sharedBuffer) {
            if (_suspended) {
                return 0;
            }
            return _size;
        }
    }

    public Id getPubId()
    {
        return _pubId;
    }

    public Object getSubKey()
    {
        return _subkey;
    }

    public boolean hasSubscriber(Id subId)
    {
        synchronized (_sharedBuffer) {
            return _policy.get(subId) != null;
        }
    }

    public void addSubscriber(Id subId, Object policy, Buffer buffer)
    {
        synchronized (_sharedBuffer) {
            _policy.put(subId, policy);
            if (buffer != null && buffer.size() > 0) {
                // this only happens when subscriber moves and causes
                // suspended buffer transfer over at subrootproto, which
                // has one-to-one mapping of subId and pubId, so the
                // group buffer must be empty at this time
                if (_sharedBuffer.size() != 0 
                    || _privateBuffers.size() > 0) {
                    throw new IllegalStateException(
                        "Illegal state for buffer transfer");
                }
                _sharedBuffer.putAll(buffer.getAll());
                _size = buffer.size();
            }
        }
    }

    public Buffer removeSubscriber(Id subId)
    {
        if (!hasSubscriber(subId)) {
            return null;
        }
        Buffer ret = null;
//        synchronized (_plock) {
        synchronized (_sharedBuffer) {
            // first get events in private buffer
            BufferHandle handle = (BufferHandle) _privateBuffers.get(subId);
            if (handle != null) {
                _size -= handle.size();
                ret = handle.removeSubscriber(subId);
            }
//        }
            ret = (ret==null) ? new Buffer() : ret;

            // then merge events from shared buffer
            ret.putAll(_sharedBuffer.getAll());
            // if this is the only subscriber, clear the whole thing
            _policy.remove(subId);
            if (_policy.size() == 0) {
                _size = 0;
                _sharedBuffer.clear();
            }            
        }
        return ret;
    }

    public void putEvent(Id pubId, Id subId, Event evt)
        throws BufferFullException
    {
        synchronized (_sharedBuffer) {
            // check whether this event is in multicast mode
            if (subId == null || subId.equals(pubId)) {
                _sharedBuffer.put(evt);
                if (isSuspended() 
                    && _sharedBuffer.size() > _suspendLimit) {
                    pack();
                }
            }
            else if (hasSubscriber(subId)) {
                putSubEvent(subId, evt);
            }
            else {
                throw new IllegalArgumentException(
                    "Invalid subscriber ID " + subId);
            }
            _size ++;
        }
    }

    public void putEvents(Id pubId, Id subId, Buffer buffer)
        throws BufferFullException
    {
        synchronized (_sharedBuffer) {
            // check whether this event is still in multicast mode
            if (subId == null || subId.equals(pubId)) {
                _sharedBuffer.putAll(buffer.getAll());
                if (isSuspended() 
                    && _sharedBuffer.size() > _suspendLimit) {
                    pack();
                }
            }
            else if (hasSubscriber(subId)) {
                putSubEvents(subId, buffer);
            }
            else {
                throw new IllegalArgumentException(
                    "Invalid subscriber ID " + subId);
            }
            _size += buffer.size();
        }
    }

    public BufferedEvent getBufferedEvent()
    {
        List ret = new Vector();
//        synchronized (_plock) {
        synchronized (_sharedBuffer) {
            if (_privateBuffers.size() > 0) {
                BufferHandle handle = null;
                // get a list of non-empty sub buffers
                Iterator itr = _privateBuffers.values().iterator();
                while (itr.hasNext()) {
                    handle = (BufferHandle) itr.next();
                    if (handle.size() > 0) {
                        ret.add(handle);
                    }
                }
                // randomly pick one buffer and get an event
                if (ret.size() > 0) {
                    handle = (BufferHandle) 
                        ret.get(_unigen.nextIntFromTo(0, ret.size()-1));
                    BufferedEvent bufevt = new BufferedEvent(
                        _pubId, (Address)_subkey, 
                        handle.getBufferedEvent().getEvent());
                    // clear the private buffers if no event left
                    if (handle.size() == 0 && ret.size() == 1) {
                        _privateBuffers.clear();
                    }
                    synchronized (_sharedBuffer) {
                        _size --;
                    }
                    return bufevt;
                }
            }
            // if no event in sub buffers, get one from group buffer
            BufferedEvent bufevt = 
                new BufferedEvent(_pubId, (Address)_subkey, _sharedBuffer.get());
            _size --;
            return bufevt;
        }
    }

    // merge all the shared events into private buffers and then pack them all 
    public void pack()
    {
        synchronized (_sharedBuffer) {
            int oldsize = _size;
//            synchronized (_plock) {
                try {
                    if (_policy.size() > 0) {
                        Iterator itr = _policy.keySet().iterator();
                        while (itr.hasNext()) {
                            Id subId = (Id) itr.next();
                            BufferHandle handle = 
                                (BufferHandle) _privateBuffers.get(subId);
                            if (handle == null) {
                                handle = new SingleBufferHandle(
                                    _pubId, subId, _suspendLimit);
                                handle.addSubscriber(
                                    subId, _policy.get(subId), null);
                                _privateBuffers.put(subId, handle);
                            }
                            int osize = handle.size();
                            handle.putEvents(_pubId, subId, _sharedBuffer);
                            handle.pack();
                            int nsize = handle.size();
                            _size += nsize - osize;
                        }
                    }
                }
                catch (Exception ex) {
                    ex.printStackTrace();
                }
//            }
            _size -= _sharedBuffer.size();
            _sharedBuffer.clear();
        }
    }

    // suspend all private buffers as well
    public void suspend()
    {
        synchronized (_sharedBuffer) {
            _suspended = true;
            synchronized (_plock) {
                if (_privateBuffers.size() > 0) {
                    Iterator itr = _privateBuffers.values().iterator();
                    while (itr.hasNext()) {
                        BufferHandle handle = (BufferHandle) itr.next();
                        handle.suspend();
                    }
                }
            }
        }
    }

    // resume all private buffers as well
    public void resume()
    {
        synchronized (_sharedBuffer) {
            _suspended = false;
            synchronized (_plock) {
                if (_privateBuffers.size() > 0) {
                    Iterator itr = _privateBuffers.values().iterator();
                    while (itr.hasNext()) {
                        BufferHandle handle = (BufferHandle) itr.next();
                        handle.resume();
                    }
                }
            }
        }
    }

    public boolean isSuspended()
    {
        synchronized (_sharedBuffer) {
            return _suspended;
        }
    }

    public String toString()
    {
        return _pubId.toString() + " -> " 
            + _subkey.toString() + " [" + size() + "]\n";
    }

    //=======================================================================
    // Private helper methods.
    //=======================================================================

    private void putSubEvent(Id subId, Event evt)
        throws BufferFullException
    {
        synchronized (_sharedBuffer)
        {
            BufferHandle handle = (BufferHandle) _privateBuffers.get(subId);
            if (handle == null) {
                handle = new SingleBufferHandle(_pubId, subId, _suspendLimit);
                handle.addSubscriber(subId, _policy.get(subId), null);
                _privateBuffers.put(subId, handle);
            }
            handle.putEvent(_pubId, subId, evt);
        }
    }

    private void putSubEvents(Id subId, Buffer buffer)
        throws BufferFullException
    {
        synchronized (_sharedBuffer)
        {
            BufferHandle handle = (BufferHandle) _privateBuffers.get(subId);
            if (handle == null) {
                handle = new SingleBufferHandle(_pubId, subId, _suspendLimit);
                handle.addSubscriber(subId, _policy.get(subId), null);
                _privateBuffers.put(subId, handle);
            }
            handle.putEvents(_pubId, subId, buffer);
        }
    }
}
