/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.pack.proto;


import solar.api.Id;
import solar.api.Event;
import solar.module.protocol.Buffer;
import solar.module.protocol.BufferFullException;
import solar.service.pack.PackPolicy;


/**
 *  Manages a single subscription.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: SingleBufferHandle.java,v 1.7 2003/10/13 19:34:26 glchen Exp $
 */
class SingleBufferHandle 
    implements BufferHandle
{
    protected Id _subId;
    protected Id _pubId;
    protected Object _subkey;
    protected Object _policy;
    protected Buffer _buffer;
    protected boolean _suspended;
    protected int _suspendLimit;
    protected BufferPacker _packer;
    
    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public SingleBufferHandle(Id pubId, Object subkey, int suspendLimit)
    {
        _pubId = pubId;
        _subkey = subkey;
        _suspended = false;
        _suspendLimit = suspendLimit;
        _buffer = new Buffer();
        _packer = BufferPackerFactory.makeBufferPacker();
    }

    //=======================================================================
    // Implementation of BufferHandle interface.
    //=======================================================================

    public synchronized int size()
    {
        if (_suspended) {
            return 0;
        }
        else {
            return _buffer.size();
        }
    }

    public Id getPubId()
    {
        return _pubId;
    }

    public Object getSubKey()
    {
        return _subkey;
    }

    public boolean hasSubscriber(Id subId)
    {
        if (_subId == null) {
            return false;
        }
        else {
            return _subId.equals(subId);
        }
    }

    public synchronized void addSubscriber(
        Id subId, Object policy, Buffer buffer)
    {
        if (_subId == null) {
            _subId = subId;
        }
        else if (!_subId.equals(subId)) {
            throw new IllegalArgumentException(
                "Invalid subscriber ID " + subId);
        }
        _policy = policy;
        _buffer = (buffer==null) ? _buffer : buffer;
    }

    public synchronized Buffer removeSubscriber(Id subId)
    {
        if (!hasSubscriber(subId)) {
            throw new IllegalArgumentException(
                "Invalid subscriber ID " + subId);
        }
        _subId = null;
        Buffer ret = _buffer;
        _buffer = new Buffer();
        return ret;
    }

    public synchronized void putEvent(Id pubId, Id subId, Event evt)
        throws BufferFullException
    {
        _buffer.put(evt);
        if (isSuspended() 
            && size() > _suspendLimit) {
            pack();
        }
    }

    public synchronized void putEvents(Id pubId, Id subId, Buffer buffer)
        throws BufferFullException
    {
        _buffer.putAll(buffer.getAll());
        if (isSuspended() 
            && size() > _suspendLimit) {
            pack();
        }
    }

    public synchronized BufferedEvent getBufferedEvent()
    {
        return new BufferedEvent(_pubId, _subId, _buffer.get());
    }

    public synchronized void pack()
    {
        try {
            _buffer = _packer.pack(_buffer, (PackPolicy) _policy);
        }
        catch (Exception ex) { 
            ex.printStackTrace(); 
        }
    }

    public synchronized void suspend()
    {
        _suspended = true;
    }

    public synchronized void resume()
    {
        _suspended = false;
    }

    public synchronized boolean isSuspended()
    {
        return _suspended;
    }

    public String toString()
    {
        return _pubId.toString() + " -> " 
            + _subId.toString() + " [" + size() + "]\n";
    }
}
