/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.rmtobj;


import serp.util.Options;
import org.apache.log4j.Logger;
import java.util.Timer;
import java.util.TimerTask;
import java.util.List;
import java.util.Vector;
import java.util.Set;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Timer;
import java.util.TimerTask;

import solar.api.Id;
import solar.api.Event;
import solar.api.Address;
import solar.api.Attribute;
import solar.api.Request;
import solar.api.RequestHandler;
import solar.impl.EventFactory;
import solar.module.cache.Cache;
import solar.module.cache.CachedObject;
import solar.module.cache.RefreshableCache;
import solar.service.BaseService;
import solar.service.ServiceException;
import solar.service.dispatch.DispatchService;
import solar.service.rmtobj.comm.PingEvent;
import solar.service.rmtobj.comm.RegisterRequest;
import solar.service.rmtobj.comm.UnregisterRequest;
import solar.service.rpc.RpcService;
import solar.service.rpc.RpcConstants;
import solar.service.rpc.RpcException;
import solar.service.transport.TransportService;


/**
 *  This RemoteObjectService uses a softstate protocol to monitoring remote
 *  object and uses RPC service to invoke request. Current implementation 
 *  only allows one monitor for a single remote object.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: SolarRemoteObjectService.java,v 1.7 2003/10/05 18:51:00 glchen Exp $
 */
public class SolarRemoteObjectService 
    extends BaseService
    implements RemoteObjectService, RemoteObjectConstants, 
        RpcConstants, Cache.CacheListener
{
    protected Logger _log = Logger.getLogger(SolarRemoteObjectService.class);
    
    protected String _rmtobjName;
    protected String _transportName;
    protected Address _transportAddress;
    
    protected RpcService _rpc;
    protected DispatchService _dispatch;

    protected Cache _cache;
    protected long _interval;
    protected Timer _timer = new Timer();
    protected Hashtable _local = new Hashtable(); // (Id, RequestHandler)
    protected Hashtable _remote = new Hashtable(); // (RemoteObject, RemoteObjectListener)
    protected Hashtable _registered = new Hashtable(); // (Address, HashSet(Id))

    
    //=======================================================================
    // Implementation of solar.service.Service interface.
    //=======================================================================

    public void init()
        throws ServiceException
    {
        // get the interval of periodic ping
        Options config = getServiceManager().getServiceConfig(getName());
        _interval = config.getLongProperty("interval", RMTOBJ_INTERVAL);

        _rmtobjName = getName();        
        _transportName = config.getProperty("transport", "transport");
    }
    
    public void initDependents()
        throws ServiceException
    {
        super.initDependents();     // must call at the top

        // get rpc and transport services, and local address
        _rpc = (RpcService) getDependentService("rpc", "rpc");
        _dispatch = 
            (DispatchService) getDependentService("dispatch", "dispatch");
        _transportAddress = _dispatch.getTransportAddress(_transportName);

        // get cache service and make cache for remote objects
        _cache = new RefreshableCache(RMTOBJ_CAPACITY, RMTOBJ_TTL);
        _cache.addListener(this);
        _timer.schedule((TimerTask)_cache, 1000, _interval);

        // create ping task and schedule it with a timer
        PingTask ptask = new PingTask();
        new Timer().schedule(ptask, 0, _interval); // ping interval
    }

    public Object handleRequest(Request request)
        throws Exception
    {
        Object ret = null;
        if (request instanceof RegisterRequest)
        {
            Id id = ((RegisterRequest)request).getId();
            synchronized (_local)
            {
                if (_local.get(id) == null)
                {
                    throw new IllegalArgumentException("Remote object not found.");
                }
            }
            Address address = request.getSourceAddress();
            synchronized (_registered)
            {
                Set ids = (Set) _registered.get(address);
                if (ids == null)
                {
                    ids = new HashSet();
                    _registered.put(address, ids);
                }
                ids.add(id);
            }
        }
        else if (request instanceof UnregisterRequest)
        {
            Id id = ((UnregisterRequest)request).getId();
            synchronized (_local)
            {
                if (_local.get(id) == null)
                {
                    throw new IllegalArgumentException("Remote object not found.");
                }
            }
            Address address = request.getSourceAddress();
            synchronized (_registered)
            {
                Set ids = (Set) _registered.get(address);
                if (ids == null)
                {
                    throw new IllegalArgumentException(
                        "Requesting address not registered before.");
                }
                ids.remove(id);
            }
        }
        else
        {
            throw new RpcException(
                "Failed to handle rmtobj request " + request + ".");
        }
        return ret;
    }

    public void handleEvent(Event evt)
        throws Exception
    {
        // special trick if we use Transport instead of ObjectTransport
        PingEvent pevt = (evt instanceof PingEvent) ? 
            (PingEvent) evt : new PingEvent(evt);
        // update all the remote objects in the cache
        Address address = pevt.getAddress();
        Iterator itr = pevt.getIds().iterator();
        while (itr.hasNext())
        {
            Id id = (Id) itr.next();
            RemoteObject ro = new RemoteObject(address, id);
            _log.debug("rmtobj updated: " + ro);
            _cache.addObject(ro, ro);
        }
    }


    //=======================================================================
    // Implementation of RemoteObjectService interface.
    //=======================================================================

    public void notifyObjectExpired(CachedObject co)
    {
        RemoteObject ro = (RemoteObject) co.getObject();
        _log.debug("rmtobj expired: " + co);
        RemoteObjectListener listener = null;
        synchronized (_remote)
        {
            listener = (RemoteObjectListener) _remote.get(co.getObject());
        }
        if (listener != null)
        {
            listener.accessFailed((RemoteObject)co.getObject());
        }
    }


    //=======================================================================
    // Implementation of RemoteObjectService interface.
    //=======================================================================
    
    public void register(RemoteObject rmtobj, RemoteObjectListener listener)
        throws RemoteObjectException
    {
        Request request = new RegisterRequest(rmtobj.getId());
        request.setSourceAddress(_transportAddress);        
        try
        {
            callRpc(rmtobj.getAddress(), request);
            synchronized (_remote)
            {
                _remote.put(rmtobj, listener);
            }
            _log.debug("rmtobj registered: " + rmtobj);
        }
        catch (Exception ex)
        {
            throw new RemoteObjectException(ex);
        }
    }

    public void unregister(RemoteObject rmtobj)
        throws RemoteObjectException
    {
        Request request = new UnregisterRequest(rmtobj.getId());
        request.setSourceAddress(_transportAddress);
        try
        {
            callRpc(rmtobj.getAddress(), request);
            synchronized (_remote)
            {
                _remote.remove(rmtobj);
            }
        }
        catch (Exception ex)
        {
            throw new RemoteObjectException(ex);
        }
    }

    public void register(Id id, RequestHandler handler)
    {
        synchronized (_local)
        {
            _local.put(id, handler);
        }
    }

    public void register(Id id)
    {
        synchronized (_local)
        {
            _local.put(id, new Object());
        }
    }

    public void unregister(Id id)
    {
        synchronized (_local)
        {
            _local.remove(id);
        }
    }

    public Object request(RemoteObject rmtobj, Request req)
        throws RemoteObjectException
    {
        try
        {
            return callRpc(rmtobj.getAddress(), req);
        }
        catch (Exception ex)
        {
            throw new RemoteObjectException(ex);
        }
    }


    //=======================================================================
    //  Private helper methods.
    //=======================================================================

    private Object callRpc(Address address, Request request)
        throws RpcException
    {
        return _rpc.invoke(address, _rmtobjName, request, RPC_TIMEOUT);
    }
    

    //=======================================================================
    //  Private helper class.
    //=======================================================================

    /** The task periodically notifies remote monitors about local objects. */
    class PingTask
        extends TimerTask
    {
        //===================================================================
        // Implementation of TimerTask interface.
        //===================================================================

        public void run()
        {
            List attr = new Vector();
            Iterator itr = _registered.keySet().iterator();
            while (itr.hasNext())
            {
                Address address = (Address) itr.next();
                Set ids = (Set) _registered.get(address);
                Event evt = new PingEvent(_transportAddress, ids);
                try
                {
                    _dispatch.dispatchEvent(evt, 
                        address, _rmtobjName, _transportName);
                }
                catch (Exception ex)
                {
                    ex.printStackTrace();
                }
            }
        }
    }
}
