/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.routing;


import serp.util.Options;
import org.apache.log4j.Logger;
import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.util.Vector;
import java.util.Iterator;
import java.util.Hashtable;
import java.util.Timer;
import java.util.TimerTask;

import solar.api.Id;
import solar.api.Event;
import solar.api.Address;
import solar.api.Attribute;
import solar.api.AttributeRecord;
import solar.impl.EventFactory;
import solar.module.cache.Cache;
import solar.module.cache.CachedObject;
import solar.module.cache.CacheException;
import solar.module.cache.RefreshableCache;
import solar.service.BaseService;
import solar.service.ServiceManager;
import solar.service.ServiceException;
import solar.service.ServiceRuntimeException;
import solar.service.dispatch.DispatchService;
import solar.service.routing.comm.RoutingEvent;
import solar.service.transport.SocketAddress;
import solar.service.transport.TransportService;


/**
 *  The routing table is manually setup as the overlay network is being 
 *  constructed.  The construction follows one simple rule: each node is
 *  added peering with an existing one.  The spanning tree resulted from
 *  such construction forms the routing table.  This service requires
 *  a transport, a cache, and a timer services.
 *  
 *  The valid properties are:
 *  
 *  <UL>
 *  <LI>interval: periodic probing interval, default as 5 seconds.<BR>
 *  <LI>bootstrap: the peer node to connect at startup.<BR>
 *  <LI>cache: the name of required cache service, default as "cache".<BR>
 *  <LI>timer: the name of required timer service, default as "timer".<BR>
 *  </UL>
 *  
 *  @author     Guanling Chen
 *  @version    $Id: ManualRoutingService.java,v 1.27 2003/10/05 18:51:01 glchen Exp $
 *  
 *  TODO:
 *      To prevent loop (maintain a spanning tree) during recovery phase.
 */
public class ManualRoutingService
    extends BaseService
    implements RoutingService, Cache.CacheListener, RoutingConstants
{
    protected Logger _log = Logger.getLogger(ManualRoutingService.class);
    
    protected Address _peer;    // the bootstrap peer node
    protected long _interval;   // periodic probing interval
    
    // nearby cache content (nearbyAddress, nearbyAddress)
    // remtoe cache content (remoteAddress, nearbyAddress) / the route
    protected Cache _nearby;   // cache id for a list of direct neighbors
    protected Cache _remote;   // cache id for a list of remote hosts

    protected String _routingName;
    protected String _transportName;
    protected Address _transportAddress;

    protected Timer _timer = new Timer();
    protected DispatchService _dispatch;


    
    //=======================================================================
    //  Implementation of solar.service.Service interface.
    //=======================================================================
    
    public void init()
        throws ServiceException
    {
        Options config = getServiceManager().getServiceConfig(getName());
        _interval = config.getLongProperty("interval", ROUTING_INTERVAL);
        _transportName = config.getProperty("transport", "transport");

        // configure peering node
        String peer = "";
        try
        {
            peer = config.getProperty("bootstrap");
            if (peer != null)
            {
                _peer = new SocketAddress(peer);
            }
        }
        catch (Exception ex)
        {
            throw new ServiceException(ex);
        }
    }

    public void initDependents()
        throws ServiceException
    {
        super.initDependents();     // must call at the top
        
        // cast transport service and local address
        _dispatch = (DispatchService) 
            getDependentService("dispatch", "dispatch");
        _transportAddress = _dispatch.getTransportAddress(_transportName);
        
        // get caches for neighbor and remote hosts
        _nearby = new RefreshableCache(20, 3*_interval);
        _remote = new RefreshableCache(200, 5*_interval);
        try
        {
            if (_peer != null)
            {
                _nearby.addObject(_peer, _peer);
                _log.debug("added peer [" + _peer + "] to host list.");
            }
            _nearby.addListener(this);
            _remote.addListener(this);
            _timer.schedule((TimerTask)_nearby, 1000, _interval);
            _timer.schedule((TimerTask)_remote, 1000, _interval);
        }
        catch (CacheException ce)
        {
            throw new ServiceException(ce);
        }

        // get timer and schedule a routing task
        RoutingTask rtask = new RoutingTask();
        Timer timer = new Timer();
        timer.schedule(rtask, 500, _interval);
    }
    
    /** A periodic routing task. */
    class RoutingTask
        extends TimerTask
    {
        //=======================================================================
        //  Implementation of java.util.TimerTask interface.
        //=======================================================================
        public synchronized void run()
        {
            try
            {
                Event evt = new RoutingEvent(_transportAddress, 
                    _transportAddress, _transportAddress);
                Iterator itr = _nearby.getObjects().iterator();
                while (itr.hasNext())
                {
                    Address dest = (Address) itr.next();
                    _dispatch.dispatchEvent(evt, 
                        dest, _routingName, _transportName);
                }
            }
            catch (Exception ex)
            {
                _log.error(ex);
                ex.printStackTrace();
            }
        }
    }


    //=======================================================================
    //  Implementation of solar.service.cache.Cache.CacheListener interface.
    //=======================================================================

    public void notifyObjectExpired(CachedObject co)
    {
        Address host = (Address) co.getObjectId();
        _log.debug("host [" + host + "] expired and removed from host list.");

        try
        {
            // in case of node upgrade from remote to nearby, don't recover
            if (_nearby.getObjectIds().contains(host))
                return;
        
            Iterator itr = _remote.getCachedObjects().iterator();
            while (itr.hasNext())
            {
                co = (CachedObject) itr.next();
                Address nearby = (Address) co.getObject();
                if (host.equals(nearby))    // a neighbor is lost, recovering
                {
                    // make its remote as new nearghbor
                    Address remote = (Address) co.getObjectId();
                    _nearby.addObject(remote, remote);
                    _log.debug("recover through new neighbor [" + remote + "].");
                    break;
                }
            }
        }
        catch (Exception ex)
        {
            _log.error(ex);
            ex.printStackTrace();
        }
    }

    
    //=======================================================================
    //  Implementation of solar.service.Service interface.
    //=======================================================================

    public synchronized void handleEvent(Event evt)
        throws Exception
    {
        // special trick if we use Transport instead of ObjectTransport
        RoutingEvent revt = (evt instanceof RoutingEvent) ? 
            (RoutingEvent) evt : new RoutingEvent(evt);

        Address source = revt.getSource();
        Address host = revt.getHost();
        Address lasthop = revt.getLastHop();
        if (host == null || source == null || lasthop == null)
            throw new Exception("routing event has missing fields.");
            
        if (_nearby.getObject(source) == null)
        {
            _log.debug("adding host [" + source + "] to nearby list.");
        }
        _nearby.addObject(source, source);

        if (_transportAddress.equals(host) 
            || _transportAddress.equals(lasthop))   // looping
        {
            /* does not work, may cause simultaneous cutting off
            if (_laddr.equals(hostAddr) 
                && _laddr.equals(preAddr))   // bigger looping
            {
                // cut off a neighbor
                _cache.removeObject(_nearby, srcAddr);
            }*/
            return;
        }

        if (_nearby.getObject(host) == null)
        {
            if ((_remote.getObject(host) == null))
            {
                _log.debug("adding host [" + host + "] to remote list.");
            }
            _remote.addObject(host, source);
        }

        // relay announcements to direct neighbors
        revt = new RoutingEvent(_transportAddress, host, source);
        Iterator itr = _nearby.getObjects().iterator();
        while (itr.hasNext())
        {
            Address dest = (Address) itr.next();
            _dispatch.dispatchEvent(revt, dest, _routingName, _transportName);
        }
    }

    
    //=======================================================================
    //  Implementation of solar.service.routing.RoutingService interface.
    //=======================================================================
    
    public synchronized Address getNextHop(Address dest)
        throws RoutingException
    {
        if (_nearby.getObjectIds().contains(dest))
            return dest;

        Address nextHop = (Address) _remote.getObject(dest);
        if (nextHop == null)
        {
            throw new RoutingException(
                "No route discovered to address [" + dest + "]");
        }
        else
        {
            return nextHop;
        }
    }
}
