/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.rpc;


import java.util.Vector;
import java.util.Hashtable;


/**
 *  A blocking object that other objects can wait on till time out
 *  or a reply is ready.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: Block.java,v 1.3 2003/07/18 02:10:19 glchen Exp $
 */
class Block 
{
    protected static Hashtable _blocks        = new Hashtable();
    protected static final Object _nullobj    = new Object();


    //=======================================================================
    // Constructor.
    //=======================================================================
    
    /** Prevent object initialization. */
    private Block()
    {
    }


    //=======================================================================
    // Public interface methods.
    //=======================================================================
    
    /** Add one block into repository. */
    public static void putBlock(Object key)
    {
        synchronized (_blocks)
        {
            _blocks.put(key, _nullobj);
        }
    }

    /** Get the block EQUALS to specified key from repository. */
    public static Object getBlock(Object key)
        throws IllegalArgumentException
    {
        synchronized (_blocks)
        {
            if (!_blocks.containsKey(key))
                throw new IllegalArgumentException(
                    "RPC block " + key + " does not exist.");

            // Retrieve the equavilant block to the argument key.
            // In many cases we cannot directly wake up the existing
            // block using network received key since they are different
            // objects. Instead, call getBlock to retrieve original
            // waiting block.
            Vector vet = new Vector(_blocks.keySet());
            int index = vet.indexOf(key);
            Object oldkey = vet.elementAt(index);

            return oldkey;
        }
    }

    /** Remove one block from repository. */
    public static void removeBlock(Object key)
    {
        synchronized (_blocks)
        {
            _blocks.remove(key);
        }
    }
    
    
    /** Wait for reply with specified timeout.  This method must be called
     *  within synchronization of the key as follows:
     *  <P>
     *  
     *  Block.putBlock(origianlKey);<BR>
     *  <BR>
     *  synchronized(origianlKey)<BR>
     *  {<BR>
     *      // do some action such as network transmission<BR>
     *      reply = Block.waitBlock(origianlKey, timeout);<BR>
     *  }<BR>
     */
    public static Object waitBlock(Object key, long timeout)
    {
        try
        {
            if (timeout == 0)
                key.wait();
            else
                key.wait(timeout); 
        }
        catch (InterruptedException ie)
        {
        }
        
        synchronized (_blocks)
        {
            Object ret = _blocks.get(key);
            if (ret.equals(_nullobj))
                return null;
            else
            {
                _blocks.remove(key);
                return ret;
            }
        }
    }

    /** Wake up the blocks waiting for replies.  This method must be
     *  called within synchronization of the key and the key must be the
     *  original waiting block retrieved by calling getBlock as follows:
     *  <P>
     *  
     *  Object originalKey = Block.getBlock(receivedKey);<BR>
     *  <BR>
     *  synchronized(originalKey)<BR>
     *  {<BR>
     *      Block.wakeBlock(originalKey, result);<BR>
     *  }<BR>
     *  */
    public static void wakeBlock(Object key, Object obj)
    {
        synchronized (_blocks)
        {
            _blocks.put(key, obj);
            key.notifyAll();
        }
    }
}
