/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.rpc;


import org.apache.log4j.Logger;
import serp.util.Options;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Hashtable;

import solar.api.Id;
import solar.api.Event;
import solar.api.Request;
import solar.api.Address;
import solar.service.BaseService;
import solar.service.ServiceManager;
import solar.service.ServiceException;
import solar.service.dispatch.DispatchUnit;
import solar.service.dispatch.DispatchService;
import solar.service.rpc.comm.ReplyEvent;
import solar.service.rpc.comm.RequestEvent;


/**
 *  An implementation of RPC service and it requires a dispatch
 *  service.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: SolarRpcService.java,v 1.16 2003/10/05 18:51:01 glchen Exp $
 */
public class SolarRpcService
    extends BaseService
    implements RpcService
{
    /** Logging facility. */
    protected static Logger _log = Logger.getLogger(SolarRpcService.class);

    /** Dependent dispatch service. */
    protected DispatchService _dispatch;

    /** The service manager. */
    protected ServiceManager _manager;

    /** The name of this RPC service. */
    protected String _rpcName;

    /** The reusable RPC acknowledge object. */
    protected static final RpcAck _rpcAck = new RpcAck();

    /** The name of transport service used. */
    protected String _transportName;

    
    //=======================================================================
    //  Implementation of solar.service.Service interface.
    //=======================================================================

    public void init()
        throws ServiceException
    {
        _rpcName = getName();
        Options config = getServiceManager().getServiceConfig(_rpcName);
        _transportName = config.getProperty("transport", "transport");
    }
    
    public void initDependents()
        throws ServiceException
    {
        super.initDependents();     // must call at the top

        _manager = getServiceManager();
        _dispatch = (DispatchService) 
            getDependentService("dispatch", "dispatch");
    }

    public void handleDispatchUnit(DispatchUnit dunit)
        throws Exception
    {
        Event evt = (Event) dunit.getAdu();
        if (evt instanceof RequestEvent)
        {
            doHandleRequest((RequestEvent) evt, dunit.getSourceAddress());
        }
        else if (evt instanceof ReplyEvent)
        {
            dohandleReply((ReplyEvent) evt);
        }
        // special tricks if we use Transport instead of ObjectTransport
        else if (RequestEvent.isRequestEvent(evt))
        {
            doHandleRequest(new RequestEvent(evt), dunit.getSourceAddress());
        }
        else if (ReplyEvent.isReplyEvent(evt))
        {
            dohandleReply(new ReplyEvent(evt));
        }
        else
        {
            _log.error("Unexpected RPC event: " + evt);
        }
    }

    // should not be called since all requests are disptached for other service
    public Object handleRequest(Request req)
        throws Exception
    {
        _log.warn("Dropping received request for rpc itself: " + req);
        throw new IllegalStateException("Not expected to handle request.");
    }


    //=======================================================================
    // Private helper methods.
    //=======================================================================

    private void dohandleReply(ReplyEvent evt)
    {
        _log.debug("Receiving RPC reply " + evt);
        
        Id block = evt.getBlock();
        try
        {
            block = (Id) Block.getBlock(block);
        }
        catch (IllegalArgumentException iae)
        {
            iae.printStackTrace();
            _log.error(iae);
            return;
        }
        
        Object ret = evt.getReply();
        synchronized (block)
        {
            Block.wakeBlock(block, ret);
        }
    }
    
    private void doHandleRequest(RequestEvent evt, Address source)
    {
        _log.debug("Receiving RPC request " + evt);

        // create a new thread to handle the request
        new RpcWorker(evt, source).start();
    }
    

    //=======================================================================
    // Implementation of solar.service.rpc.RpcService interface.
    //=======================================================================
    
    public Object invoke(Address dest, String handler, Request req)
        throws RpcException
    {
        return invoke(dest, handler, req, 0);
    }

    public Object invoke(
        Address dest, String handler, Request req, long timeout)
        throws RpcException
    {
        assertInited();

        // create a block for this request
        Id block = Id.makeId("block-");
        Block.putBlock(block);

        // send request to destination as an event
        RequestEvent re = new RequestEvent(req, handler, block);
        Object reply;
        synchronized (block)
        {
            try
            {
                _dispatch.dispatchEvent(re, dest, _rpcName, _transportName);
            }
            catch (Exception ex)
            {
                throw new RpcException("Failed to send RPC request.", ex);
            }
        
            // get result object
            reply = Block.waitBlock(block, timeout);
        }
        
        if (reply == null)
        {
            throw new RpcException("Timed out for RPC request.");
        }
        
        if (reply instanceof Throwable)
        {
            throw new RpcException(
                "Remote exception for RPC request.", 
                (Throwable) reply);
        }

        reply = (reply instanceof RpcAck) ? null : reply;
        return reply;
    }



    /**  A worker thread to handle one RPC request. */
    class RpcWorker 
        extends Thread
    {
        protected Address _source;
        protected RequestEvent _evt;
    
        //===================================================================
        // Contructors.
        //===================================================================
    
        public RpcWorker(RequestEvent evt, Address source)
        {
            super("rpc-worker");
            _evt = evt;
            _source = source;
        }

        //===================================================================
        // Public interface methods.
        //===================================================================

        public void run()
        {
            // retrieve data fields from RPC request event
            Id block = _evt.getBlock();
            String handler = _evt.getHandler();
            Request req = _evt.getRequest();

            // call the handling service to service the request
            // exception will be captured as return result
            Object ret = null;
            try
            {
                ret = _manager.getService(handler).handleRequest(req);
            }
            catch (Exception ex)
            {
                // restrict the exception length, to prevent UDP overflow
                StringWriter sw = new StringWriter();
                PrintWriter pw = new PrintWriter(sw);
                pw.flush();
                ex.printStackTrace(pw);
                String msg = sw.toString();
                if (msg.length() > 500)
                {
                    msg = msg.substring(0, 500);
                }
                ret = new Exception(msg);
            }

            // if the handling service returns null, force the result to be an
            // empty RpcAck object.
            ret = (ret == null) ? _rpcAck : ret;

            // construct RPC reply event
            ReplyEvent re = new ReplyEvent(ret, block);
            try
            {
                _dispatch.dispatchEvent(re, _source, _rpcName, _transportName);
            }
            catch (Exception ex)
            {
                ex.printStackTrace();
                _log.error("Failed to send RPC reply.", ex);
            }
        }
    }
}
