/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.transport;


import org.apache.log4j.Logger;
import java.util.List;
import java.util.Vector;
import java.util.Iterator;

import rice.pastry.Id;
import rice.pastry.PastryNode;
import rice.pastry.client.CommonAPIAppl;
import rice.pastry.security.Credentials;
import rice.pastry.security.PermissiveCredentials;

import solar.api.Adu;
import solar.api.Address;
import solar.api.AduFactory;
import solar.impl.ByteArrayAdu;


/**
 *  A Pastry transport application interfacing with Pastry routing substrate.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: PastryTransport.java,v 1.12 2003/10/05 18:51:03 glchen Exp $
 */
public class PastryTransport
    extends CommonAPIAppl
    implements Transport
{
    /** Logging facility. */
    protected static Logger _log = Logger.getLogger(PastryTransport.class);
    
    /** Local address, multiplexing over same Pastry node. */
    protected static rice.pastry.messaging.Address _maddr = new MultiplexAddress();

    /** Credentials of this Pastry application / our transport layer. */
    protected static Credentials _cred = new PermissiveCredentials();

    /** The pastry node associated with this application. */
    protected PastryNode _pnode;

    /** The list of transport listeners. */
    protected List _listeners = new Vector();
    
    /** The local transport address for Pastry node. */
    protected Address _laddr;

    protected AduFactory _aduFactory;


    //=======================================================================
    //  Private helper class.
    //=======================================================================
    
    private static class MultiplexAddress 
        implements rice.pastry.messaging.Address
    {
        private int myCode = 0x1984abcd;
	
        public int hashCode() { return myCode; }

        public boolean equals(Object obj) 
        {
            return (obj instanceof MultiplexAddress);
        }

        public String toString() { return "[TransportAddress]"; }
    }


    //=======================================================================
    //  Constructors.
    //=======================================================================
    
    public PastryTransport(PastryNode pn)
    {
        super(pn);
        _pnode = pn;
        _laddr = (Address) new PastryAddress(getNodeId());
        _log.info("Pastry transport started at local address: " + _laddr);
    }


    //=======================================================================
    //  Public interface methods.
    //=======================================================================

    public PastryNode getPastryNode()
    {
        return _pnode;
    }

    
    //=======================================================================
    // Implementation of solar.service.transport.Transport interface.
    //=======================================================================

    public void setAduFactory(AduFactory aduFactory)
    {
        _aduFactory = aduFactory;
    }
    
    public Address getLocalAddress()
    {
        return _laddr;
    }

    public Address getAddressByKey(String key)
        throws TransportException
    {
        return new PastryAddress(key.getBytes());
    }
    
    public void send(byte[] data, Address dest)
        throws TransportException
    {
        send(data, dest, null);
    }

    public void send(byte[] data, Address dest, solar.api.Id multiplex)
        throws TransportException
    {
        Id destNodeId = ((PastryAddress)dest).getNodeId();
        PastryMessage msg = new PastryMessage(data, _maddr);

        Header header = new Header(multiplex);
        header.setSource(_laddr);
        header.setDestination(dest);
        msg.setHeader(header);
        
    	route(destNodeId, msg, null);
    }

    public void send(Adu adu, Address dest)
        throws TransportException
    {
        send(adu, dest, null);
    }

    public void send(Adu adu, Address dest, solar.api.Id multiplex)
        throws TransportException
    {
        send(adu.toBytes(), dest, multiplex);
    }

    public void addTransportListener(TransportListener listener)
    {
        _listeners.add(listener);
    }

    public void removeTransportListener(TransportListener listener)
    {
        _listeners.remove(listener);
    }

    public void stop()
    {
        // does nothing
    }


    //=======================================================================
    // Private helper methods.
    //=======================================================================
    
    /** Notify transport listeners about received data. */
    protected void notifyDataReceived(Adu adu, Header header)
    {
        Iterator itr = _listeners.iterator();
        while (itr.hasNext())
        {
            TransportListener tl = (TransportListener) itr.next();
            tl.dataReceived(adu, header);
        }
    }


    //=======================================================================
    //  Implementation of rice.pastry.client.CommonAPIAppl interface.
    //=======================================================================

    public rice.pastry.messaging.Address getAddress()
    {
        return _maddr;
    }

    public Credentials getCredentials()
    {
        return _cred;
    }

    public void deliver(Id key, rice.pastry.messaging.Message msg)
    {
        Adu adu;
        PastryMessage p = (PastryMessage) msg;
        Header header = p.getHeader();

        try
        {
            byte[] data = p.getBytes();
            if (_aduFactory != null)
            {
                adu = _aduFactory.parse(data);
            }
            else
            {
                adu = new ByteArrayAdu(data);
            }
            notifyDataReceived(adu, header);
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
            _log.error(ex);
        }
    }
}
