/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.transport;


import serp.util.Options;
import org.apache.log4j.Logger;
import java.net.InetAddress;
import java.net.InetSocketAddress;

import rice.pastry.PastryNode;
import rice.pastry.standard.IPNodeIdFactory;
import rice.pastry.dist.DistPastryNodeFactory;
import rice.pastry.dist.NewDistPastryNodeFactory;
import solar.service.ServiceException;


/**
 *  Transport service using Pastry routing substrate.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: PastryTransportService.java,v 1.24 2003/10/12 00:58:39 glchen Exp $
 */
public class PastryTransportService
    extends BaseTransportService
    implements TransportConstants
{
    protected int _port = TRANSPORT_PASTRY_PORT;
    protected String _protocol = null;
    protected boolean _debug = false;
    protected boolean _serialize = true;

    
    //=======================================================================
    //  Implementation of solar.service.Service interface.
    //=======================================================================
    
    public void init()
        throws ServiceException
    {
        _log = Logger.getLogger(PastryTransportService.class);
        
        try
        {
            Options config = getServiceManager().getServiceConfig(getName());
            _port = config.getIntProperty("port", TRANSPORT_PASTRY_PORT);
            _protocol = config.getProperty("protocol", "wire");
            _debug = config.getBooleanProperty("debug", false);
            _serialize = config.getBooleanProperty("serialize", true);
            if (_debug)
            {
                String[] args = new String[1];
                args[0] = "-verbose";
                rice.pastry.Log.init(args);
            }
            String bootstrap = config.getProperty("bootstrap");
            if (bootstrap == null)
            {
                _transport = getPastryTransport(null);
            }
            else
            {
                bootstrap = (bootstrap.indexOf(":") == -1) ? 
                    bootstrap + ":" + Integer.toString(TRANSPORT_PASTRY_PORT) 
                    : bootstrap;
                _transport = getPastryTransport(new SocketAddress(bootstrap));
            }
            _transport.addTransportListener(this);
            if (!_serialize)
            {
                _transport.setAduFactory(_aduFactory);
            }
            _log.debug("got transport: " + _transport);
        }
        catch (ServiceException se)
        {
            throw se;
        }
        catch (Exception ex)
        {
            throw new ServiceException(ex);
        }
    }


    //=======================================================================
    //  Private helper methods.
    //=======================================================================

    /** Return the instance of Pastry transport layer. */
    private PastryTransport getPastryTransport(SocketAddress bootstrap)
        throws Exception
    {
        _log.debug("Pastry bootstrap: " + bootstrap);

        
        // create Pastry ID and node factories
        IPNodeIdFactory idFactory = new IPNodeIdFactory(_port);
        DistPastryNodeFactory factory = null;
        if (_protocol.equals("wire"))
        {
            factory = DistPastryNodeFactory.getFactory(
                idFactory, 
                DistPastryNodeFactory.PROTOCOL_WIRE, 
                _port);
        }
        else if (_protocol.equals("rmi"))
        {
            factory = DistPastryNodeFactory.getFactory(
                idFactory, 
                DistPastryNodeFactory.PROTOCOL_RMI, 
                _port);            
        }
        else if (_protocol.equals("tcp"))
        {
            factory = NewDistPastryNodeFactory.getFactory(
                _serialize,
                idFactory, 
                NewDistPastryNodeFactory.PROTOCOL_TCP, 
                _port);
        }
        else if (_protocol.equals("udp"))
        {
            factory = NewDistPastryNodeFactory.getFactory(
                _serialize,
                idFactory, 
                NewDistPastryNodeFactory.PROTOCOL_UDP, 
                _port);
        }
        else
        {
            throw new IllegalArgumentException(
                "Unsupported Pastry protocol " + _protocol);
        }

        // create the Pastry node from boostrap or by itself            
        PastryNode node = null;
        InetSocketAddress address = null;
        if(bootstrap != null)
        {
            address = new InetSocketAddress(
                bootstrap.getInetAddress(), bootstrap.getPort());
            node = factory.newNode(factory.getNodeHandle(address));
        }
        else
        {
            node = factory.newNode(null);
        }

        // return the Pastry transport substrate
        if (_serialize)
        {
            return new PastryObjectTransport(node);
        }
        else
        {
            return new PastryTransport(node);
        }
    }

}
