/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.transport;


import org.apache.log4j.Logger;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.DatagramPacket;
import java.util.Iterator;

import solar.api.Id;
import solar.api.Adu;
import solar.api.Address;
import solar.impl.ByteArrayAdu;


/**
 *  The UDP transport layer using object (ADU) serialization.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: UdpObjectTransport.java,v 1.4 2003/10/12 00:58:40 glchen Exp $
 */
public class UdpObjectTransport
    extends UdpTransport
    implements TransportConstants
{
    //=======================================================================
    // Constructors.
    //=======================================================================

    /** Constructs a new UDP transport, which listens on an available port. */
    public UdpObjectTransport()
        throws Exception
    {
        super();
        _log = Logger.getLogger(UdpObjectTransport.class);
    }

    
    /** Constructs a new UDP transport, which listens on a specified port. */
    public UdpObjectTransport(int port)
        throws Exception
    {
        super(port);
        _log = Logger.getLogger(UdpObjectTransport.class);
    }

    public UdpObjectTransport(String hostname, int port)
        throws Exception
    {
        super(hostname, port);
        _log = Logger.getLogger(UdpObjectTransport.class);
    }

    //=======================================================================
    // Implementation of solar.service.transport.Transport interface.
    //=======================================================================

    public void send(byte[] data, Address dest)
        throws TransportException
    {
        send(data, dest, null);
    }

    public void send(byte[] data, Address dest, Id multiplex)
        throws TransportException
    {
        send(new ByteArrayAdu(data), dest, multiplex);
    }

    public void send(Adu adu, Address dest)
        throws TransportException
    {
        send(adu, dest, null);
    }

    public void send(Adu adu, Address dest, Id multiplex)
        throws TransportException
    {
        try
        {
            // set up header info
            Header header = new Header(multiplex);
            header.setSource(_laddr);
            header.setDestination(dest);
            // write header and adu into bytes
            ByteArrayOutputStream bout = new ByteArrayOutputStream();
            ObjectOutputStream out = new ObjectOutputStream(bout);
            header.write(out);
            out.writeObject(adu);
            // format bytes into packet and send it
            byte[] data = bout.toByteArray();
            SocketAddress rdest = (SocketAddress) dest;
            DatagramPacket dp = 
                new DatagramPacket(data, data.length, 
                rdest.getInetAddress(), rdest.getPort());
            _ssock.send(dp);

//            System.out.println("sent " + dp.getLength() + " bytes");
        
            bout.close();
            out.close();
        }
        catch (Exception ex)
        {
            throw new TransportException("Failed to send UDP packet.", ex);
        }
    }

    //=======================================================================
    // Private helper methods.
    //=======================================================================

    /** Notify transport listeners about received message. */
    private void notifyDataReceived(Adu adu, Header header)
    {
        if (_listeners == null)
        {
            return;
        }
        Iterator itr = _listeners.iterator();
        while (itr.hasNext())
        {
            TransportListener l = (TransportListener) itr.next();
            l.dataReceived(adu, header);
        }
    }


    //=======================================================================
    //  Implementation of java.lang.Runnable interface.
    //=======================================================================

    public void run()
    {
        byte[] ret = new byte[TRANSPORT_UDP_MSS];
        DatagramPacket dp = new DatagramPacket(ret, TRANSPORT_UDP_MSS);
        while (true)
        {
            try
            {
                // receive packet and get the data buffer
                _ssock.receive(dp);
                byte[] data = dp.getData();
                // read in the header info
                ObjectInputStream in = 
                    new ObjectInputStream(
                    new ByteArrayInputStream(dp.getData()));
                Header header = new Header();
                header.read(in);
                // read in the ADU
                Adu adu = (Adu) in.readObject();
                in.close();
                // notify the listeners
                notifyDataReceived(adu, header);

            }
            catch (Exception ex)
            {
                _log.error(ex);
                ex.printStackTrace();
            }
        }
    }
}
