/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package versus;


import serp.util.Options;
import java.io.*;
import java.util.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

import solar.api.Event;
import solar.api.Request;
import solar.api.Attribute;
import solar.service.dan.Operator;
import solar.impl.EventFactory;


/**
 *  A transformer converts badge and sensor ID numbers in Versus events 
 *  into symbolic values.
 *  
 *  <p>The published events contains following fields:
 *  <ul>
 *  <li>name        (object name)
 *  <li>zone        (current location if it is an IR event, or 
 *                   radio\<col_num:sen_num\> if it is RF event.)
 *  <li>button      (whether button is pressed)
 *  <li>battery     (whether battery is still good)
 *  <li>motion      (whether the badge is moving)
 *  <li>tcount      (a rotating event counter for a badge)
 *  <li>timestamp   (when the event is received)
 *  </ul>
 *  
 *  @author     Guanling Chen
 *  @version    $Id: VersusTransformer.java,v 1.4 2003/09/19 00:38:32 glchen Exp $
 */
public class VersusTransformer 
    extends Operator
{
    private Hashtable _floorplan;
    private Hashtable _badges;


    //=======================================================================
    // Implementation of solar.service.dan.Operator interface.
    //=======================================================================
    
    /**
     *  Init the operator's floorplan and badge mapping tables
     *  specified in an XML file.
     */
    public void init(Options opts)
        throws Exception
    {
        String filename = opts.getProperty("map");
        XmlVersusParser parser = new XmlVersusParser();
        parser.parse(new File(filename));
        _floorplan = parser.getFloorplan();
        _badges = parser.getBadges();
    }


    /** Convert tag and sensor ID numbers to symbolic values. */
    public void handleEvent(Event evt) 
        throws Exception
    {
        String badge = (String) evt.getValue("badge_num");
        String collector = (String) evt.getValue("col_num");
        String sensor = (String) evt.getValue("sen_num");

        String name, zone;

        name = (String) (_badges.get(badge));
        if (name == null)
            name = "unknown_badge<" + badge + ">";

        Hashtable collectorTable = (Hashtable)_floorplan.get(collector);
        if (collectorTable == null)
        {
            zone = "unknown_collector<" + collector +">";
        }
        else 
        {
            zone = (String)(collectorTable.get(sensor));
            if (zone == null)
                zone = "unknown_sensor<" + collector + ":" + sensor +">";
        }

        if (zone.indexOf("radio") != -1)
            zone = zone + "<" + collector + ":" + sensor + ">";

        List at = new Vector();
        at.add(new Attribute("name", name));
        at.add(new Attribute("zone", zone));
        at.add(new Attribute("timestamp", evt.getValue("timestamp")));
        at.add(new Attribute("motion", evt.getValue("motion")));
        at.add(new Attribute("battery", evt.getValue("battery")));
        at.add(new Attribute("button", evt.getValue("button")));
        
        Event new_evt = EventFactory.getPlainEvent(at);
        publish(new_evt);
    }


    public Object handleQuery(Request query)
        throws Exception
    {
        return null;
    }

    
    //=======================================================================
    // Main entrance for test purpose.
    //=======================================================================
    
    /** 
     *  Method used for standalone test, not used when in Solar
     *  Operator mode.
     */
    public static void main(String[] args)
        throws Exception
    {
        Options opts = new Options();
        opts.setProperty("map", 
            "/u/solar/public_html/internal/versus/versus.xml");
        new VersusTransformer().init(opts);
    }
}
